from __future__ import annotations

import posixpath
import warnings
from abc import ABC
from typing import (
    TYPE_CHECKING,
    Any,
    Hashable,
    Iterable,
    Iterator,
    List,
    Protocol,
    Sequence,
    SupportsIndex,
    TypeVar,
    overload,
)

from .context import Context
from .paginator import Paginator

if TYPE_CHECKING:
    from .context import Context


class BaseResource(dict):
    def __init__(self, ctx: Context, /, **kwargs):
        super().__init__(**kwargs)
        self._ctx = ctx

    def __getattr__(self, name):
        if name in self:
            warnings.warn(
                f"Accessing the field '{name}' via attribute is deprecated and will be removed in v1.0.0. "
                f"Please use __getitem__ (e.g., {self.__class__.__name__.lower()}['{name}']) for field access instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return self[name]
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def update(self, *args, **kwargs):
        super().update(*args, **kwargs)


class Resources:
    def __init__(self, ctx: Context) -> None:
        self._ctx = ctx


class Active(ABC, BaseResource):
    def __init__(self, ctx: Context, path: str, /, **attributes):
        """A dict abstraction for any HTTP endpoint that returns a singular resource.

        Extends the `Resource` class and provides additional functionality for via the session context and an optional parent resource.

        Parameters
        ----------
        ctx : Context
            The context object containing the session and URL for API interactions.
        path : str
            The HTTP path component for the resource endpoint
        **attributes : dict
            Resource attributes passed
        """
        super().__init__(ctx, **attributes)
        self._ctx = ctx
        self._path = path


class Resource(Protocol):
    _ctx: Context
    _path: str

    def __getitem__(self, key: Hashable, /) -> Any: ...


class _Resource(dict, Resource):
    def __init__(self, ctx: Context, path: str, **attributes):
        self._ctx = ctx
        self._path = path
        super().__init__(**attributes)

    def destroy(self) -> None:
        self._ctx.client.delete(self._path)

    def update(self, **attributes):  # pyright: ignore[reportIncompatibleMethodOverride]
        response = self._ctx.client.put(self._path, json=attributes)
        result = response.json()
        super().update(**result)


T = TypeVar("T", bound=Resource)


class ResourceSequence(Protocol[T]):
    @overload
    def __getitem__(self, index: SupportsIndex, /) -> T: ...

    @overload
    def __getitem__(self, index: slice, /) -> List[T]: ...

    def __len__(self) -> int: ...

    def __iter__(self) -> Iterator[T]: ...

    def __str__(self) -> str: ...

    def __repr__(self) -> str: ...


class _ResourceSequence(Sequence[T], ResourceSequence[T]):
    def __init__(self, ctx: Context, path: str, *, uid: str = "guid"):
        self._ctx = ctx
        self._path = path
        self._uid = uid

    def __getitem__(self, index):
        return list(self.fetch())[index]

    def __len__(self) -> int:
        return len(list(self.fetch()))

    def __iter__(self) -> Iterator[T]:
        return iter(self.fetch())

    def __str__(self) -> str:
        return str(self.fetch())

    def __repr__(self) -> str:
        return repr(self.fetch())

    def create(self, **attributes: Any) -> Any:
        response = self._ctx.client.post(self._path, json=attributes)
        result = response.json()
        uid = result[self._uid]
        path = posixpath.join(self._path, uid)
        return _Resource(self._ctx, path, **result)

    def fetch(self, **conditions) -> Iterable[Any]:
        response = self._ctx.client.get(self._path, params=conditions)
        results = response.json()
        resources = []
        for result in results:
            uid = result[self._uid]
            path = posixpath.join(self._path, uid)
            resource = _Resource(self._ctx, path, **result)
            resources.append(resource)

        return resources

    def find(self, *args: str) -> Any:
        path = posixpath.join(self._path, *args)
        response = self._ctx.client.get(path)
        result = response.json()
        return _Resource(self._ctx, path, **result)

    def find_by(self, **conditions) -> Any | None:
        """
        Find the first record matching the specified conditions.

        There is no implied ordering, so if order matters, you should specify it yourself.

        Parameters
        ----------
        **conditions : Any

        Returns
        -------
        Optional[T]
            The first record matching the conditions, or `None` if no match is found.
        """
        collection = self.fetch(**conditions)
        return next((v for v in collection if v.items() >= conditions.items()), None)


class _PaginatedResourceSequence(_ResourceSequence):
    def fetch(self, **conditions):
        paginator = Paginator(self._ctx, self._path, dict(**conditions))
        for page in paginator.fetch_pages():
            resources = []
            results = page.results
            for result in results:
                uid = result[self._uid]
                path = posixpath.join(self._path, uid)
                resource = _Resource(self._ctx, path, **result)
                resources.append(resource)
            yield from resources



import pytest
import responses
from responses import matchers

from posit.connect.client import Client
from posit.connect.content import ContentItem
from posit.connect.resources import _Resource

from .api import load_mock, load_mock_dict


class TestContentItemGetContentOwner:
    @responses.activate
    def test_owner(self):
        mock_content = load_mock_dict("v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json")
        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=mock_content,
        )

        mock_user_get = responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )

        c = Client("https://connect.example", "12345")
        item = c.content.get("f2f37341-e21d-3d80-c698-a935ad614066")
        owner = item.owner
        assert owner["guid"] == "20a79ce3-6e87-4522-9faf-be24228800a4"

        # load a second time, assert tha owner is loaded from cached result
        owner = item.owner
        assert owner["guid"] == "20a79ce3-6e87-4522-9faf-be24228800a4"
        assert mock_user_get.call_count == 1


class TestContentItemDelete:
    @responses.activate
    def test(self):
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_get = responses.get(
            f"https://connect.example/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )

        mock_delete = responses.delete(f"https://connect.example/__api__/v1/content/{guid}")

        # setup
        c = Client("https://connect.example", "12345")
        content = c.content.get(guid)

        # invoke
        content.delete()

        # assert
        assert mock_delete.call_count == 1


class TestContentDeploy:
    @responses.activate
    def test(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"
        task_id = "jXhOhdm5OOSkGhJw"

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_content_deploy = responses.post(
            f"https://connect.example/__api__/v1/content/{content_guid}/deploy",
            match=[matchers.json_params_matcher({"bundle_id": None})],
            json={"task_id": task_id},
        )

        mock_tasks_get = responses.get(
            f"https://connect.example/__api__/v1/tasks/{task_id}",
            json=load_mock(f"v1/tasks/{task_id}.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")
        content = c.content.get(content_guid)

        # invoke
        task = content.deploy()

        # assert
        assert task["id"] == task_id
        assert mock_content_get.call_count == 1
        assert mock_content_deploy.call_count == 1
        assert mock_tasks_get.call_count == 1


class TestContentUpdate:
    @responses.activate
    def test_update(self):
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        responses.get(
            f"https://connect.example/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )
        con = Client("https://connect.example", "12345")
        content = con.content.get(guid)
        assert content["guid"] == guid

        new_name = "New Name"
        fake_content = load_mock_dict(f"v1/content/{guid}.json")
        fake_content.update(name=new_name)
        responses.patch(
            f"https://connect.example/__api__/v1/content/{guid}",
            json=fake_content,
        )

        content.update(name=new_name)
        assert content["name"] == new_name


class TestContentCreate:
    @responses.activate
    def test(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fake_content_item = load_mock_dict(f"v1/content/{guid}.json")

        # behavior
        responses.post(
            "https://connect.example/__api__/v1/content",
            json=load_mock(f"v1/content/{guid}.json"),
            match=[matchers.json_params_matcher({"name": fake_content_item["name"]})],
        )

        # setup
        client = Client(api_key="12345", url="https://connect.example/")

        # invoke
        fake_name = fake_content_item["name"]
        assert isinstance(fake_name, str)
        content_item = client.content.create(name=fake_name)

        # assert
        assert content_item["name"] == fake_content_item["name"]


class TestContentsFind:
    @responses.activate
    def test(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        content = client.content.find()

        # assert
        assert mock_get.call_count == 1
        assert len(content) == 3
        assert content[0]["name"] == "team-admin-dashboard"
        assert content[1]["name"] == "Performance-Data-1671216053560"
        assert content[2]["name"] == "My-Streamlit-app"

    @responses.activate
    def test_params_include(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[matchers.query_param_matcher({"include": "tags"})],
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        client.content.find(include="tags")

        #  assert
        assert mock_get.call_count == 1

    @responses.activate
    def test_params_include_list(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[matchers.query_param_matcher({"include": "tags,owner"})],
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        client.content.find(include=["tags", "owner"])

        #  assert
        assert mock_get.call_count == 1

    @responses.activate
    def test_params_include_none(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[matchers.query_param_matcher({})],
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        client.content.find(include=None)

        #  assert
        assert mock_get.call_count == 1


class TestContentsFindBy:
    @responses.activate
    def test(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        content = client.content.find_by(name="team-admin-dashboard")

        # assert
        assert mock_get.call_count == 1
        assert content
        assert content["name"] == "team-admin-dashboard"

    @responses.activate
    def test_miss(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        content = client.content.find_by(name="does-not-exist")

        # assert
        assert mock_get.call_count == 1
        assert content is None


class TestContentsFindOne:
    @responses.activate
    def test(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        content_item = client.content.find_one()

        #  assert
        assert mock_get.call_count == 1
        assert content_item

    @responses.activate
    def test_owner_guid(self):
        # data
        owner_guid = "a01792e3-2e67-402e-99af-be04a48da074"

        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        content_item = client.content.find_one(owner_guid=owner_guid)

        #  assert
        assert mock_get.call_count == 1
        assert content_item
        assert content_item["owner_guid"] == owner_guid

    @responses.activate
    def test_name(self):
        # data
        name = "team-admin-dashboard"

        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        content_item = client.content.find_one(name=name)

        #  assert
        assert mock_get.call_count == 1
        assert content_item
        assert content_item["name"] == name

    @responses.activate
    def test_params_include(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[matchers.query_param_matcher({"include": "tags"})],
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        client.content.find_one(include="tags")

        #  assert
        assert mock_get.call_count == 1

    @responses.activate
    def test_params_include_none(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
            match=[matchers.query_param_matcher({})],
        )

        # setup
        client = Client("https://connect.example", "12345")

        # invoke
        client.content.find_one(include=None)

        #  assert
        assert mock_get.call_count == 1


class TestContentsGet:
    @responses.activate
    def test(self):
        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=load_mock("v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json"),
        )
        con = Client("https://connect.example", "12345")
        get_one = con.content.get("f2f37341-e21d-3d80-c698-a935ad614066")
        assert get_one["name"] == "Performance-Data-1671216053560"


class TestContentsCount:
    @responses.activate
    def test(self):
        responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )
        con = Client(api_key="12345", url="https://connect.example/")
        count = con.content.count()
        assert count == 3


class TestRender:
    @responses.activate
    def test(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        get_content = responses.get(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )

        patch_content = responses.patch(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )

        get_variants = responses.get(
            f"https://connect.example.com/__api__/applications/{guid}/variants",
            json=load_mock(f"applications/{guid}/variants.json"),
        )

        post_render = responses.post(
            "https://connect.example.com/__api__/variants/6627/render",
            json=load_mock("variants/6627/render.json"),
        )

        # setup
        c = Client("https://connect.example.com", "12345")
        content = c.content.get(guid)

        # invoke
        task = content.render()

        # assert
        assert task is not None
        assert get_content.call_count == 1
        assert patch_content.call_count == 1
        assert get_variants.call_count == 1
        assert post_render.call_count == 1

    @responses.activate
    def test_app_mode_is_other(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fixture_content = load_mock_dict(f"v1/content/{guid}.json")
        fixture_content.update(app_mode="other")

        # behavior
        responses.get(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=fixture_content,
        )

        responses.patch(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=fixture_content,
        )

        # setup
        c = Client("https://connect.example.com", "12345")
        content = c.content.get(guid)

        # invoke
        with pytest.raises(ValueError):
            content.render()

    @responses.activate
    def test_missing_default(self):
        responses.get(
            "https://connect.example.com/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=load_mock("v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json"),
        )

        responses.patch(
            "https://connect.example.com/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=load_mock("v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json"),
        )

        responses.get(
            "https://connect.example.com/__api__/applications/f2f37341-e21d-3d80-c698-a935ad614066/variants",
            json=[],
        )

        c = Client("https://connect.example.com", "12345")
        content = c.content.get("f2f37341-e21d-3d80-c698-a935ad614066")
        with pytest.raises(RuntimeError):
            content.render()


class TestRestart:
    @responses.activate
    def test(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fixture_content = load_mock_dict(f"v1/content/{guid}.json")
        fixture_content.update(app_mode="api")

        # behavior
        mock_get_content = mock_get = responses.get(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=fixture_content,
        )

        mock_patch_content = responses.patch(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=fixture_content,
        )

        mock_patch_environment = responses.patch(
            f"https://connect.example.com/__api__/v1/content/{guid}/environment",
        )

        mock_get_content_page = responses.get(
            f"https://connect.example.com/content/{guid}",
        )

        # setup
        c = Client("https://connect.example.com", "12345")
        content = c.content.get(guid)

        # invoke
        task = content.restart()

        # assert
        assert task is None
        assert mock_get_content.call_count == 1
        assert mock_patch_content.call_count == 1
        assert mock_patch_environment.call_count == 2
        assert mock_get_content_page.call_count == 1

    @responses.activate
    def test_app_mode_is_other(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fixture_content = load_mock_dict(f"v1/content/{guid}.json")
        fixture_content.update(app_mode="other")

        # behavior
        mock_get_content = mock_get = responses.get(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=fixture_content,
        )

        mock_patch_content = responses.patch(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=fixture_content,
        )

        # setup
        c = Client("https://connect.example.com", "12345")
        content = c.content.get(guid)

        # invoke
        with pytest.raises(ValueError):
            content.restart()

        # assert
        assert mock_get_content.call_count == 1
        assert mock_patch_content.call_count == 1


class TestContentRepository:
    base_url = "http://connect.example"
    client = Client(base_url, "12345")

    @property
    def content_guid(self):
        return "f2f37341-e21d-3d80-c698-a935ad614066"

    @property
    def content_item(self):
        return ContentItem(self.ctx, guid=self.content_guid)

    @property
    def endpoint(self):
        return f"{self.base_url}/__api__/v1/content/{self.content_guid}/repository"

    @property
    def ctx(self):
        return self.client._ctx

    def mock_repository_info(self):
        content_item = self.content_item

        mock_get = responses.get(
            self.endpoint,
            json=load_mock_dict(f"v1/content/{self.content_guid}/repository.json"),
        )
        repository_info = content_item.repository

        assert isinstance(repository_info, _Resource)
        assert mock_get.call_count == 1

        return repository_info

    @responses.activate
    def test_repository_getter_returns_repository(self):
        # Performs assertions in helper property method
        self.mock_repository_info()

    @responses.activate
    def test_repository_update(self):
        repository_info = self.mock_repository_info()

        mock_patch = responses.patch(
            self.endpoint,
            json=load_mock_dict(f"v1/content/{self.content_guid}/repository_patch.json"),
        )
        repository_info.update(branch="testing-main")
        assert mock_patch.call_count == 1

        for key, value in repository_info.items():
            if key == "branch":
                assert repository_info[key] == "testing-main"
            else:
                assert repository_info[key] == value

    @responses.activate
    def test_repository_delete(self):
        repository_info = self.mock_repository_info()

        mock_delete = responses.delete(self.endpoint)
        repository_info.destroy()

        assert mock_delete.call_count == 1



"""Tag resources."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional, overload

from typing_extensions import NotRequired, TypedDict, Unpack

from ._utils import update_dict_values
from .context import Context, ContextManager
from .resources import Active

if TYPE_CHECKING:
    from .content import ContentItem


class _RelatedTagsBase(ContextManager, ABC):
    @abstractmethod
    def find(self) -> list[Tag]:
        pass


def _update_parent_kwargs(kwargs: dict) -> dict:
    """
    Sets the `parent_id` key in the kwargs if `parent` is provided.

    Asserts that the `parent=` and `parent_id=` keys are not both provided.
    """
    parent = kwargs.get("parent", None)
    parent_id = kwargs.get("parent_id", None)
    if parent is None:
        # No parent to upgrade, return the kwargs as is
        return kwargs
    if parent and parent_id:
        raise ValueError("Cannot provide both `parent=` and `parent_id=`")
    if not isinstance(parent, Tag):
        raise TypeError(
            "`parent=` must be a Tag instance. If using a string, please use `parent_id=`"
        )

    # Remove `parent` from a copy of `kwargs` and replace it with `parent_id`
    ret_kwargs = {**kwargs}
    del ret_kwargs["parent"]
    ret_kwargs["parent_id"] = parent["id"]

    return ret_kwargs


class Tag(Active):
    """Tag resource."""

    class _Attrs(TypedDict, total=False):
        id: str
        """The identifier for the tag."""
        name: str
        """The name of the tag."""
        parent_id: NotRequired[Optional[str]]
        """The identifier for the parent tag. If there is no parent_id, the tag is a top-level tag."""
        created_time: str
        """The timestamp (RFC3339) indicating when the tag was created. Ex. '2006-01-02T15:04:05Z'"""
        updated_time: str
        """The timestamp (RFC3339) indicating when the tag was created. Ex. '2006-01-02T15:04:05Z'"""

    def __init__(self, ctx: Context, path: str, /, **kwargs: Unpack[Tag._Attrs]):
        super().__init__(ctx, path, **kwargs)

    @property
    def parent_tag(self) -> Tag | None:
        if self.get("parent_id", None) is None:
            return None

        parent = self._ctx.client.tags.get(self["parent_id"])
        return parent

    @property
    def child_tags(self) -> ChildTags:
        """
        Find all child tags that are direct children of this tag.

        Returns
        -------
        ChildrenTags
            Helper class that can `.find()` the child tags.

        Examples
        --------
        ```python
        import posit

        client = posit.connect.Client()
        mytag = client.tags.find(id="TAG_ID_HERE")

        children = mytag.child_tags.find()
        ```
        """
        return ChildTags(self._ctx, self._path, parent_tag=self)

    @property
    def descendant_tags(self) -> DescendantTags:
        """
        Find all tags that descend from this tag.

        Returns
        -------
        DescendantTags
            Helper class that can `.find()` all descendant tags.

        Examples
        --------
        ```python
        import posit

        client = posit.connect.Client()
        mytag = client.tags.find(id="TAG_ID_HERE")

        descendant_tags = mytag.descendant_tags.find()
        ```
        """
        return DescendantTags(self._ctx, parent_tag=self)

    @property
    def content_items(self) -> TagContentItems:
        """
        Find all content items that are tagged with this tag.

        Returns
        -------
        TagContentItems
            Helper class that can `.find()` all content items.

        Examples
        --------
        ```python
        import posit

        client = posit.connect.Client()
        first_tag = client.tags.find()[0]

        first_tag_content_items = first_tag.content_items.find()
        ```
        """
        path = f"{self._path}/content"
        return TagContentItems(self._ctx, path)

    def destroy(self) -> None:
        """
        Removes the tag.

        Deletes a tag, including all descendants in its own tag hierarchy.

        Examples
        --------
        ```python
        import posit

        client = posit.connect.Client()
        first_tag = client.tags.find()[0]

        # Remove the tag
        first_tag.destroy()
        ```
        """
        self._ctx.client.delete(self._path)

    # Allow for every combination of `name` and (`parent` or `parent_id`)
    @overload
    def update(self, /, *, name: str = ..., parent: Tag | None = ...) -> None: ...
    @overload
    def update(self, /, *, name: str = ..., parent_id: str | None = ...) -> None: ...

    def update(
        self,
        **kwargs,
    ) -> None:
        """
        Update the tag.

        Parameters
        ----------
        name : str
            The name of the tag.
        parent : Tag | None, optional
            The parent `Tag` object. If there is no parent, the tag is a top-level tag. To remove
            the parent tag, set the value to `None`. Only one of `parent` or `parent_id` can be
            provided.
        parent_id : str | None, optional
            The identifier for the parent tag. If there is no parent, the tag is a top-level tag.
            To remove the parent tag, set the value to `None`.

        Returns
        -------
        Tag
            Updated tag object.

        Examples
        --------
        ```python
        import posit

        client = posit.connect.Client()
        last_tag = client.tags.find()[-1]

        # Update the tag's name
        updated_tag = last_tag.update(name="new_name")

        # Remove the tag's parent
        updated_tag = last_tag.update(parent=None)
        updated_tag = last_tag.update(parent_id=None)

        # Update the tag's parent
        parent_tag = client.tags.find()[0]
        updated_tag = last_tag.update(parent=parent_tag)
        updated_tag = last_tag.update(parent_id=parent_tag["id"])
        ```
        """
        updated_kwargs = _update_parent_kwargs(kwargs)
        response = self._ctx.client.patch(self._path, json=updated_kwargs)
        result = response.json()
        update_dict_values(self, **result)


class TagContentItems(ContextManager):
    def __init__(self, ctx: Context, path: str) -> None:
        super().__init__()
        self._ctx = ctx
        self._path = path

    def find(self) -> list[ContentItem]:
        """
        Find all content items that are tagged with this tag.

        Returns
        -------
        list[ContentItem]
            List of content items that are tagged with this tag.

        Examples
        --------
        ```python
        import posit

        client = posit.connect.Client()
        first_tag = client.tags.find()[0]

        first_tag_content_items = first_tag.content_items.find()
        ```
        """
        from .content import ContentItem

        response = self._ctx.client.get(self._path)
        results = response.json()
        return [ContentItem(self._ctx, **result) for result in results]


class ChildTags(_RelatedTagsBase):
    def __init__(self, ctx: Context, path: str, /, *, parent_tag: Tag) -> None:
        super().__init__()
        self._ctx: Context = ctx
        self._path: str = path

        self._parent_tag = parent_tag

    def find(self) -> list[Tag]:
        """
        Find all child tags that are direct children of a single tag.

        Returns
        -------
        list[Tag]
            List of child tags. (Does not include the parent tag.)

        Examples
        --------
        ```python
        import posit

        client = posit.connect.Client()
        mytag = client.tags.get("TAG_ID_HERE")

        child_tags = mytag.child_tags.find()
        ```
        """
        child_tags = self._ctx.client.tags.find(parent=self._parent_tag)
        return child_tags


class DescendantTags(_RelatedTagsBase):
    def __init__(self, ctx: Context, /, *, parent_tag: Tag) -> None:
        super().__init__()
        self._ctx = ctx
        self._parent_tag = parent_tag

    def find(self) -> list[Tag]:
        """
        Find all child tags that descend from a single tag.

        Returns
        -------
        list[Tag]
            List of tags that descend from the parent tag.
        """
        # This method could be done using `tags.find(parent=self._root_id)` but it would require
        # a request for every child tag recursively. (O(n) requests)
        # By using the `/v1/tags` endpoint, we can get all tags in a single request
        # and filter them in Python. (1 request)

        all_tags = self._ctx.client.tags.find()

        # O(n^2) algorithm to find all child tags. O(n) in practice.
        #
        # This could be optimized by using a dictionary to store the tags by their parent_id and
        # then recursively traverse the dictionary to find all child tags. O(2 * n) = O(n) but the
        # code is more complex
        #
        # If the tags are always ordered (which they seem to be ordered by creation date - parents are first),
        # this algo be performed in a two passes (one to add all tags, one to confirm no more additions)
        child_tags = []
        parent_ids = {self._parent_tag["id"]}
        tag_found: bool = True
        while tag_found:
            tag_found = False
            for tag in [*all_tags]:
                parent_id = tag.get("parent_id")
                if not parent_id:
                    # Skip tags with no parent
                    all_tags.remove(tag)
                    continue
                if parent_id in parent_ids:
                    # Child found, remove from search list
                    child_tags.append(tag)
                    parent_ids.add(tag["id"])
                    # Remove from search list
                    all_tags.remove(tag)
                    tag_found = True

        return child_tags


class Tags(ContextManager):
    """Content item tags resource."""

    def __init__(self, ctx: Context, path: str) -> None:
        super().__init__()
        self._ctx = ctx
        self._path = path

    def _tag_path(self, tag_id: str) -> str:
        if not isinstance(tag_id, str):
            raise TypeError('Tag `"id"` must be a string')
        if tag_id == "":
            raise ValueError('Tag `"id"` cannot be an empty string')

        return f"{self._path}/{tag_id}"

    def get(self, tag_id: str) -> Tag:
        """
        Get a single tag by its identifier.

        Parameters
        ----------
        tag_id : str
            The identifier for the tag.

        Returns
        -------
        Tag
            The tag object.

        Examples
        --------
        ```python
        import posit

        client = posit.connect.Client()
        mytag = client.tags.get("TAG_ID_HERE")
        ```
        """
        if not isinstance(tag_id, str):
            raise TypeError("`tag_id` must be a string")

        if tag_id == "":
            raise ValueError("`tag_id` cannot be an empty string")
        path = self._tag_path(tag_id)
        response = self._ctx.client.get(path)
        return Tag(self._ctx, path, **response.json())

    # Allow for every combination of `name` and (`parent` or `parent_id`)
    @overload
    def find(self, /, *, name: str = ..., parent: Tag = ...) -> list[Tag]: ...
    @overload
    def find(self, /, *, name: str = ..., parent_id: str = ...) -> list[Tag]: ...

    def find(self, /, **kwargs) -> list[Tag]:
        """
        Find tags by name and/or parent.

        Note: tag names are only unique within the scope of a parent, which means that it is
        possible to have multiple results when querying by name; However, querying by both `name`
        and `parent` ensures a single result.

        Parameters
        ----------
        name : str, optional
            The name of the tag.
        parent : Tag, optional
            The parent `Tag` object. If there is no parent, the tag is a top-level tag. Only one of
            `parent` or `parent_id` can be provided.
        parent_id : str, optional
            The identifier for the parent tag. If there is no parent, the tag is a top-level tag.

        Returns
        -------
        list[Tag]
            List of tags that match the query. Defaults to all Tags.

        Examples
        --------
        ```python
        import posit

        client = posit.connect.Client()

        # Find all tags
        all_tags = client.tags.find()

        # Find all tags with the name
        mytag = client.tags.find(name="tag_name")

        # Find all tags with the name and parent
        subtags = client.tags.find(name="sub_name", parent=mytag)
        subtags = client.tags.find(name="sub_name", parent=mytag["id"])
        ```
        """
        updated_kwargs = _update_parent_kwargs(
            kwargs,  # pyright: ignore[reportArgumentType]
        )
        response = self._ctx.client.get(self._path, params=updated_kwargs)
        results = response.json()
        return [Tag(self._ctx, self._tag_path(result["id"]), **result) for result in results]

    @overload
    def create(self, /, *, name: str) -> Tag: ...
    @overload
    def create(self, /, *, name: str, parent: Tag) -> Tag: ...
    @overload
    def create(self, /, *, name: str, parent_id: str) -> Tag: ...

    def create(self, /, **kwargs) -> Tag:
        """
        Create a tag.

        Parameters
        ----------
        name : str
            The name of the tag.
        parent : Tag, optional
            The parent `Tag` object. If there is no parent, the tag is a top-level tag. Only one of
            `parent` or `parent_id` can be provided.
        parent_id : str, optional
            The identifier for the parent tag. If there is no parent, the tag is a top-level tag.

        Returns
        -------
        Tag
            Newly created tag object.

        Examples
        --------
        ```python
        import posit

        client = posit.connect.Client()

        category_tag = client.tags.create(name="category_name")
        tag = client.tags.create(name="tag_name", parent=category_tag)
        ```
        """
        updated_kwargs = _update_parent_kwargs(
            kwargs,  # pyright: ignore[reportArgumentType]
        )

        response = self._ctx.client.post(self._path, json=updated_kwargs)
        result = response.json()
        return Tag(self._ctx, self._tag_path(result["id"]), **result)


class ContentItemTags(ContextManager):
    """Content item tags resource."""

    def __init__(self, ctx: Context, path: str, /, *, tags_path: str, content_guid: str) -> None:
        super().__init__()
        self._ctx = ctx

        # v1/content/{guid}/tags
        self._path = path
        self._tags_path = tags_path

        self._content_guid = content_guid

    def find(self) -> list[Tag]:
        """
        Find all tags that are associated with a single content item.

        Returns
        -------
        list[Tag]
            List of tags associated with the content item.

        Examples
        --------
        ```python
        import posit

        client = posit.connect.Client()
        content_item = client.content.find_one()

        # Find all tags associated with the content item
        content_item_tags = content_item.tags.find()
        ```
        """
        response = self._ctx.client.get(self._path)
        results = response.json()

        tags: list[Tag] = []
        for result in results:
            tags.append(
                Tag(
                    self._ctx,
                    f"{self._tags_path}/{result['id']}",
                    **result,
                )
            )

        return tags

    def _to_tag_id(self, tag: str | Tag) -> str:
        if isinstance(tag, Tag):
            tag_id = tag["id"]
            if not isinstance(tag_id, str):
                raise TypeError(f'Expected `tag=` `"id"` to be a string. Received: {tag}')

        elif isinstance(tag, str):
            tag_id = tag
        else:
            raise TypeError(f"Expected `tag=` to be a string or Tag object. Received: {tag}")

        if tag_id == "":
            raise ValueError(f"Expected 'tag=' ID to be non-empty. Received: {tag_id}")

        return tag_id

    def add(self, tag: str | Tag) -> None:
        """
        Add the specified tag to an individual content item.

        When adding a tag, all tags above the specified tag in the tag tree are also added to the
        content item.

        Parameters
        ----------
        tag : str | Tag
            The tag id or tag object to add to the content item.

        Examples
        --------
        ```python
        import posit

        client = posit.connect.Client()

        content_item = client.content.find_one()
        tag = client.tags.find()[0]

        # Add a tag
        content_item.tags.add(tag)
        ```
        """
        tag_id = self._to_tag_id(tag)

        self._ctx.client.post(self._path, json={"tag_id": tag_id})

    def delete(self, tag: str | Tag) -> None:
        """
        Remove the specified tag from an individual content item.

        When removing a tag, all tags above the specified tag in the tag tree are also removed from
        the content item.

        Parameters
        ----------
        tag : str | Tag
            The tag id or tag object to remove from the content item.

        Examples
        --------
        ```python
        import posit

        client = posit.connect.Client()

        content_item = client.content.find_one()
        content_item_first_tag = content_item.tags.find()[0]

        # Remove a tag
        content_item.tags.delete(content_item_first_tag)
        ```
        """
        tag_id = self._to_tag_id(tag)

        self._ctx.client.delete(f"{self._path}/{tag_id}")



import responses
from responses import matchers

from posit.connect.client import Client

from ..api import load_mock, load_mock_dict


class TestIntegrationDelete:
    @responses.activate
    def test(self):
        guid = "22644575-a27b-4118-ad06-e24459b05126"

        # behavior
        responses.get(
            f"https://connect.example/__api__/v1/oauth/integrations/{guid}",
            json=load_mock(f"v1/oauth/integrations/{guid}.json"),
        )

        mock_delete = responses.delete(
            f"https://connect.example/__api__/v1/oauth/integrations/{guid}",
        )

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None
        integration = c.oauth.integrations.get(guid)

        # invoke
        integration.delete()

        # assert
        assert mock_delete.call_count == 1


class TestIntegrationUpdate:
    @responses.activate
    def test(self):
        # data
        guid = "22644575-a27b-4118-ad06-e24459b05126"
        responses.get(
            f"https://connect.example/__api__/v1/oauth/integrations/{guid}",
            json=load_mock(f"v1/oauth/integrations/{guid}.json"),
        )

        c = Client("https://connect.example", "12345")
        c._ctx.version = None
        integration = c.oauth.integrations.get(guid)
        assert integration["guid"] == guid

        new_name = "New Name"

        fake_integration = load_mock_dict(f"v1/oauth/integrations/{guid}.json")
        fake_integration.update(name=new_name)
        assert fake_integration["name"] == new_name

        mock_update = responses.patch(
            f"https://connect.example/__api__/v1/oauth/integrations/{guid}",
            json=fake_integration,
        )

        integration.update(name=new_name)
        assert mock_update.call_count == 1
        assert integration["name"] == new_name


class TestIntegrationsCreate:
    @responses.activate
    def test(self):
        # data
        guid = "22644575-a27b-4118-ad06-e24459b05126"
        fake_integration = load_mock_dict(f"v1/oauth/integrations/{guid}.json")

        # behavior
        mock_create = responses.post(
            "https://connect.example/__api__/v1/oauth/integrations",
            json=load_mock(f"v1/oauth/integrations/{guid}.json"),
            match=[
                matchers.json_params_matcher(
                    {
                        "name": fake_integration["name"],
                        "description": fake_integration["description"],
                        "template": fake_integration["template"],
                        "config": fake_integration["config"],
                    },
                ),
            ],
        )

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None

        # invoke
        integration = c.oauth.integrations.create(
            name=fake_integration["name"],
            description=fake_integration["description"],
            template=fake_integration["template"],
            config=fake_integration["config"],
        )

        # assert
        assert mock_create.call_count == 1
        assert integration["name"] == fake_integration["name"]
        assert integration["description"] == fake_integration["description"]
        assert integration["template"] == fake_integration["template"]
        assert integration["config"] == fake_integration["config"]


class TestIntegrationsFind:
    @responses.activate
    def test(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/oauth/integrations",
            json=load_mock("v1/oauth/integrations.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None

        # invoke
        integrations = c.oauth.integrations.find()

        # assert
        assert mock_get.call_count == 1
        assert len(integrations) == 2
        assert integrations[0]["id"] == "3"
        assert integrations[1]["id"] == "4"


class TestIntegrationsGet:
    @responses.activate
    def test(self):
        guid = "22644575-a27b-4118-ad06-e24459b05126"

        # behavior
        mock_get = responses.get(
            f"https://connect.example/__api__/v1/oauth/integrations/{guid}",
            json=load_mock(f"v1/oauth/integrations/{guid}.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None
        integration = c.oauth.integrations.get(guid)

        assert mock_get.call_count == 1
        assert integration["guid"] == guid



import pytest
import responses
from responses import matchers

from posit.connect import Client

from .api import load_mock


@responses.activate
def test__delitem__():
    # data
    guid = "f2f37341-e21d-3d80-c698-a935ad614066"

    # behavior
    mock_get = responses.get(
        f"https://connect.example.com/__api__/v1/content/{guid}",
        json=load_mock(f"v1/content/{guid}.json"),
    )

    mock_patch = responses.patch(
        f"https://connect.example.com/__api__/v1/content/{guid}/environment",
        json=[],
        match=[
            matchers.json_params_matcher(
                [
                    {
                        "name": "TEST",
                        "value": None,
                    },
                ],
            ),
        ],
    )

    # setup
    c = Client("https://connect.example.com", "12345")
    content = c.content.get(guid)

    # invoke
    del content.environment_variables["TEST"]

    # assert
    assert mock_get.call_count == 1
    assert mock_patch.call_count == 1


@responses.activate
def test__getitem__():
    # data
    guid = "f2f37341-e21d-3d80-c698-a935ad614066"

    # behavior
    responses.get(
        f"https://connect.example.com/__api__/v1/content/{guid}",
        json=load_mock(f"v1/content/{guid}.json"),
    )

    responses.patch(
        f"https://connect.example.com/__api__/v1/content/{guid}/environment",
        json=[],
        match=[
            matchers.json_params_matcher(
                [
                    {
                        "name": "TEST",
                        "value": None,
                    },
                ],
            ),
        ],
    )

    # setup
    c = Client("https://connect.example.com", "12345")
    content = c.content.get(guid)

    # invoke
    with pytest.raises(NotImplementedError):
        content.environment_variables["TEST"]


@responses.activate
def test__iter__():
    # data
    guid = "f2f37341-e21d-3d80-c698-a935ad614066"

    # behavior
    mock_get_content = responses.get(
        f"https://connect.example.com/__api__/v1/content/{guid}",
        json=load_mock(f"v1/content/{guid}.json"),
    )

    mock_get_environment = responses.get(
        f"https://connect.example.com/__api__/v1/content/{guid}/environment",
        json=["TEST"],
    )

    # setup
    c = Client("https://connect.example.com", "12345")
    content = c.content.get(guid)

    # invoke
    iterator = iter(content.environment_variables)

    # assert
    assert next(iterator) == "TEST"
    with pytest.raises(StopIteration):
        next(iterator)

    assert mock_get_content.call_count == 1
    assert mock_get_environment.call_count == 1


@responses.activate
def test__len__():
    # data
    guid = "f2f37341-e21d-3d80-c698-a935ad614066"

    # behavior
    mock_get_content = responses.get(
        f"https://connect.example.com/__api__/v1/content/{guid}",
        json=load_mock(f"v1/content/{guid}.json"),
    )

    mock_get_environment = responses.get(
        f"https://connect.example.com/__api__/v1/content/{guid}/environment",
        json=["TEST"],
    )

    # setup
    c = Client("https://connect.example.com", "12345")
    content = c.content.get(guid)

    # invoke
    length = len(content.environment_variables)

    # assert
    assert length == 1
    assert mock_get_content.call_count == 1
    assert mock_get_environment.call_count == 1


@responses.activate
def test__setitem__():
    # data
    guid = "f2f37341-e21d-3d80-c698-a935ad614066"

    # behavior
    mock_get = responses.get(
        f"https://connect.example.com/__api__/v1/content/{guid}",
        json=load_mock(f"v1/content/{guid}.json"),
    )

    mock_patch = responses.patch(
        f"https://connect.example.com/__api__/v1/content/{guid}/environment",
        json=["TEST"],
        match=[
            matchers.json_params_matcher(
                [
                    {
                        "name": "TEST",
                        "value": "TEST",
                    },
                ],
            ),
        ],
    )

    # setup
    c = Client("https://connect.example.com", "12345")
    content = c.content.get(guid)

    # invoke
    content.environment_variables["TEST"] = "TEST"

    # assert
    assert mock_get.call_count == 1
    assert mock_patch.call_count == 1


@responses.activate
def test_clear():
    # data
    guid = "f2f37341-e21d-3d80-c698-a935ad614066"

    # behavior
    mock_get = responses.get(
        f"https://connect.example.com/__api__/v1/content/{guid}",
        json=load_mock(f"v1/content/{guid}.json"),
    )

    mock_put = responses.put(
        f"https://connect.example.com/__api__/v1/content/{guid}/environment",
        json=[],
        match=[matchers.json_params_matcher([])],
    )

    # setup
    c = Client("https://connect.example.com", "12345")
    content = c.content.get(guid)

    # invoke
    content.environment_variables.clear()

    # assert
    assert mock_get.call_count == 1
    assert mock_put.call_count == 1


@responses.activate
def test_create():
    # data
    guid = "f2f37341-e21d-3d80-c698-a935ad614066"

    # behavior
    mock_get = responses.get(
        f"https://connect.example.com/__api__/v1/content/{guid}",
        json=load_mock(f"v1/content/{guid}.json"),
    )

    mock_patch = responses.patch(
        f"https://connect.example.com/__api__/v1/content/{guid}/environment",
        json=["TEST"],
        match=[
            matchers.json_params_matcher(
                [
                    {
                        "name": "TEST",
                        "value": "TEST",
                    },
                ],
            ),
        ],
    )

    # setup
    c = Client("https://connect.example.com", "12345")
    content = c.content.get(guid)

    # invoke
    content.environment_variables.create("TEST", "TEST")

    # assert
    assert mock_get.call_count == 1
    assert mock_patch.call_count == 1


@responses.activate
def test_delete():
    # data
    guid = "f2f37341-e21d-3d80-c698-a935ad614066"

    # behavior
    mock_get = responses.get(
        f"https://connect.example.com/__api__/v1/content/{guid}",
        json=load_mock(f"v1/content/{guid}.json"),
    )

    mock_patch = responses.patch(
        f"https://connect.example.com/__api__/v1/content/{guid}/environment",
        json=[],
        match=[
            matchers.json_params_matcher(
                [
                    {
                        "name": "TEST",
                        "value": None,
                    },
                ],
            ),
        ],
    )

    # setup
    c = Client("https://connect.example.com", "12345")
    content = c.content.get(guid)

    # invoke
    content.environment_variables.delete("TEST")

    # assert
    assert mock_get.call_count == 1
    assert mock_patch.call_count == 1


@responses.activate
def test_find():
    # data
    guid = "f2f37341-e21d-3d80-c698-a935ad614066"

    # behavior
    mock_get_content = responses.get(
        f"https://connect.example.com/__api__/v1/content/{guid}",
        json=load_mock(f"v1/content/{guid}.json"),
    )

    mock_get_environment = responses.get(
        f"https://connect.example.com/__api__/v1/content/{guid}/environment",
        json=["TEST"],
    )

    # setup
    c = Client("https://connect.example.com", "12345")
    content = c.content.get(guid)

    # invoke
    envvars = content.environment_variables.find()

    # assert
    assert envvars == ["TEST"]
    assert mock_get_content.call_count == 1
    assert mock_get_environment.call_count == 1


@responses.activate
def test_items():
    # data
    guid = "f2f37341-e21d-3d80-c698-a935ad614066"

    # behavior
    mock_get_content = responses.get(
        f"https://connect.example.com/__api__/v1/content/{guid}",
        json=load_mock(f"v1/content/{guid}.json"),
    )

    mock_get_environment = responses.get(
        f"https://connect.example.com/__api__/v1/content/{guid}/environment",
        json=["TEST"],
    )

    # setup
    c = Client("https://connect.example.com", "12345")
    content = c.content.get(guid)

    # invoke
    with pytest.raises(NotImplementedError):
        content.environment_variables.items()


class TestUpdate:
    @responses.activate
    def test(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_get = responses.get(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )

        mock_patch = responses.patch(
            f"https://connect.example.com/__api__/v1/content/{guid}/environment",
            json=load_mock(f"v1/content/{guid}.json"),
            match=[
                matchers.json_params_matcher(
                    [
                        {
                            "name": "TEST",
                            "value": "TEST",
                        },
                    ],
                ),
            ],
        )

        # setup
        c = Client("https://connect.example.com", "12345")
        content = c.content.get(guid)

        # invoke
        content.environment_variables.update(TEST="TEST")

        # assert
        assert mock_get.call_count == 1
        assert mock_patch.call_count == 1

    @responses.activate
    def test_other_is_mapping(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_get = responses.get(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )

        mock_patch = responses.patch(
            f"https://connect.example.com/__api__/v1/content/{guid}/environment",
            json=load_mock(f"v1/content/{guid}.json"),
            match=[
                matchers.json_params_matcher(
                    [
                        {
                            "name": "TEST",
                            "value": "TEST",
                        },
                    ],
                ),
            ],
        )

        # setup
        c = Client("https://connect.example.com", "12345")
        content = c.content.get(guid)

        # invoke
        content.environment_variables.update({"TEST": "TEST"})

        # assert
        assert mock_get.call_count == 1
        assert mock_patch.call_count == 1

    @responses.activate
    def test_other_hasattr_keys(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_get = responses.get(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )

        mock_patch = responses.patch(
            f"https://connect.example.com/__api__/v1/content/{guid}/environment",
            json=load_mock(f"v1/content/{guid}.json"),
            match=[
                matchers.json_params_matcher(
                    [
                        {
                            "name": "TEST",
                            "value": "TEST",
                        },
                    ],
                ),
            ],
        )

        # setup
        c = Client("https://connect.example.com", "12345")
        content = c.content.get(guid)

        class Test:
            def __getitem__(self, key):
                return "TEST"

            def keys(self):
                return ["TEST"]

        # invoke
        content.environment_variables.update(Test())

        # assert
        assert mock_get.call_count == 1
        assert mock_patch.call_count == 1

    @responses.activate
    def test_other_is_iterable(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_get = responses.get(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )

        mock_patch = responses.patch(
            f"https://connect.example.com/__api__/v1/content/{guid}/environment",
            json=load_mock(f"v1/content/{guid}.json"),
            match=[
                matchers.json_params_matcher(
                    [
                        {
                            "name": "TEST",
                            "value": "TEST",
                        },
                    ],
                ),
            ],
        )

        # setup
        c = Client("https://connect.example.com", "12345")
        content = c.content.get(guid)

        # invoke
        content.environment_variables.update([("TEST", "TEST")])

        # assert
        assert mock_get.call_count == 1
        assert mock_patch.call_count == 1

    @responses.activate
    def test_other_is_iterable_of_something_else(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_get = responses.get(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )

        mock_patch = responses.patch(
            f"https://connect.example.com/__api__/v1/content/{guid}/environment",
            json=load_mock(f"v1/content/{guid}.json"),
            match=[
                matchers.json_params_matcher(
                    [
                        {
                            "name": "TEST",
                            "value": "TEST",
                        },
                    ],
                ),
            ],
        )

        # setup
        c = Client("https://connect.example.com", "12345")
        content = c.content.get(guid)

        # invoke
        with pytest.raises(TypeError):
            content.environment_variables.update([0, 1, 2, 3, 4, 5])

    @responses.activate
    def test_other_is_str(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_get = responses.get(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )

        # setup
        c = Client("https://connect.example.com", "12345")
        content = c.content.get(guid)

        # invoke
        with pytest.raises(ValueError):
            content.environment_variables.update("TEST")

    @responses.activate
    def test_other_is_bytes(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_get = responses.get(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )

        # setup
        c = Client("https://connect.example.com", "12345")
        content = c.content.get(guid)

        # invoke
        with pytest.raises(TypeError):
            content.environment_variables.update(b"TEST")

    @responses.activate
    def test_other_is_something_else(self):
        # data
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_get = responses.get(
            f"https://connect.example.com/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )

        # setup
        c = Client("https://connect.example.com", "12345")
        content = c.content.get(guid)

        # invoke
        with pytest.raises(TypeError):
            content.environment_variables.update(0)



"""OAuth integration resources."""

from typing import List, Optional, overload

from ..resources import BaseResource, Resources
from .associations import IntegrationAssociations


class Integration(BaseResource):
    """OAuth integration resource."""

    @property
    def associations(self) -> IntegrationAssociations:
        return IntegrationAssociations(self._ctx, integration_guid=self["guid"])

    def delete(self) -> None:
        """Delete the OAuth integration."""
        path = f"v1/oauth/integrations/{self['guid']}"
        self._ctx.client.delete(path)

    @overload
    def update(
        self,
        *args,
        name: str = ...,
        description: str = ...,
        config: dict = ...,
        **kwargs,
    ) -> None:
        """Update the OAuth integration.

        Parameters
        ----------
        name: str, optional
        description: str, optional
        config: dict, optional
        """

    @overload
    def update(self, *args, **kwargs) -> None:
        """Update the OAuth integration."""

    def update(self, *args, **kwargs) -> None:
        """Update the OAuth integration."""
        body = dict(*args, **kwargs)
        path = f"v1/oauth/integrations/{self['guid']}"
        response = self._ctx.client.patch(path, json=body)
        super().update(**response.json())


class Integrations(Resources):
    """Integrations resource."""

    @overload
    def create(
        self,
        *,
        name: str,
        description: Optional[str],
        template: str,
        config: dict,
    ) -> Integration:
        """Create an OAuth integration.

        Parameters
        ----------
        name : str
        description : Optional[str]
        template : str
        config : dict

        Returns
        -------
        Integration
        """

    @overload
    def create(self, **kwargs) -> Integration:
        """Create an OAuth integration.

        Returns
        -------
        Integration
        """

    def create(self, **kwargs) -> Integration:
        """Create an OAuth integration.

        Parameters
        ----------
        name : str
        description : Optional[str]
        template : str
        config : dict

        Returns
        -------
        Integration
        """
        path = "v1/oauth/integrations"
        response = self._ctx.client.post(path, json=kwargs)
        return Integration(self._ctx, **response.json())

    def find(self) -> List[Integration]:
        """Find OAuth integrations.

        Returns
        -------
        List[Integration]
        """
        path = "v1/oauth/integrations"
        response = self._ctx.client.get(path)
        return [
            Integration(
                self._ctx,
                **result,
            )
            for result in response.json()
        ]

    def get(self, guid: str) -> Integration:
        """Get an OAuth integration.

        Parameters
        ----------
        guid: str

        Returns
        -------
        Integration
        """
        path = f"v1/oauth/integrations/{guid}"
        response = self._ctx.client.get(path)
        return Integration(self._ctx, **response.json())



import pytest

from posit.connect import urls


class TestCreate:
    def test(self):
        url = "http://example.com/__api__"
        assert urls.Url(url) == url

    def test_append_path(self):
        assert urls.Url("http://example.com/") == "http://example.com/__api__"

    def test_missing_scheme(self):
        with pytest.raises(ValueError):
            urls.Url("example.com")

    def test_missing_netloc(self):
        with pytest.raises(ValueError):
            urls.Url("http://")


class TestAppend:
    def test(self):
        url = "http://example.com/__api__"
        url = urls.Url(url)
        assert url + "path" == "http://example.com/__api__/path"

    def test_slash_prefix(self):
        url = "http://example.com/__api__"
        url = urls.Url(url)
        assert url + "/path" == "http://example.com/__api__/path"

    def test_slash_suffix(self):
        url = "http://example.com/__api__"
        url = urls.Url(url)
        assert url + "path/" == "http://example.com/__api__/path"



import responses

from posit.connect.client import Client

from .api import load_mock


class TestPackagesFindBy:
    @responses.activate
    def test(self):
        mock_get = responses.get(
            "https://connect.example/__api__/v1/packages",
            json=load_mock("v1/packages.json"),
        )

        c = Client("https://connect.example", "12345")
        c._ctx.version = None

        package = c.packages.find_by(name="posit")
        assert package
        assert package["name"] == "posit"
        assert mock_get.call_count == 1



import random
import uuid

import pytest
import responses
from responses import matchers

from posit import connect
from posit.connect.client import Client
from posit.connect.context import Context
from posit.connect.groups import Group
from posit.connect.permissions import Permission, Permissions
from posit.connect.users import User

from .api import load_mock, load_mock_dict, load_mock_list


class TestPermissionDestroy:
    @responses.activate
    def test(self):
        # data
        uid = "94"
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_delete = responses.delete(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions/{uid}",
        )

        # setup
        c = connect.Client("https://connect.example", "12345")
        fake_permission = load_mock_dict(f"v1/content/{content_guid}/permissions/{uid}.json")
        permission = Permission(c._ctx, **fake_permission)

        # invoke
        permission.destroy()

        # assert
        assert mock_delete.call_count == 1


class TestPermissionUpdate:
    @responses.activate
    def test_request_shape(self):
        # test data
        uid = random.randint(0, 100)
        content_guid = str(uuid.uuid4())
        principal_guid = str(uuid.uuid4())
        principal_type = "principal_type"
        role = "role"
        extraneous = "extraneous"

        # define api behavior
        responses.put(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions/{uid}",
            json={
                # doesn't matter for this test
            },
            match=[
                # assertion
                matchers.json_params_matcher(
                    {
                        # validate that initial permission fields are set
                        "principal_guid": principal_guid,
                        "principal_type": principal_type,
                        "role": role,
                        # validate that arguments passed to update are set
                        "extraneous": extraneous,
                    },
                ),
            ],
        )

        # setup
        c = connect.Client("https://connect.example", "12345")
        permission = Permission(
            c._ctx,
            id=uid,
            content_guid=content_guid,
            principal_guid=principal_guid,
            principal_type=principal_type,
            role=role,
        )

        # invoke
        # assertion occurs in match above
        permission.update(extraneous=extraneous)

    @responses.activate
    def test_role_update(self):
        # test data
        old_role = "owner"
        new_role = "viewer"

        uid = "94"
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fake_permission = load_mock_dict(f"v1/content/{content_guid}/permissions/{uid}.json")
        fake_permission.update(role=new_role)

        # define api behavior
        uid = random.randint(0, 100)
        content_guid = str(uuid.uuid4())
        responses.put(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions/{uid}",
            json=fake_permission,
            match=[
                matchers.json_params_matcher(
                    {
                        "principal_guid": None,
                        "principal_type": None,
                        "role": new_role,
                    },
                ),
            ],
        )

        # setup
        c = connect.Client("https://connect.example", "12345")
        permission = Permission(c._ctx, id=uid, content_guid=content_guid, role=old_role)

        # assert role change with respect to api response
        assert permission["role"] == old_role
        permission.update(role=new_role)
        assert permission["role"] == new_role


class TestPermissionsCount:
    @responses.activate
    def test(self):
        # test data
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fake_permissions = load_mock_list(f"v1/content/{content_guid}/permissions.json")

        # define api behavior
        responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions",
            json=fake_permissions,
        )

        # setup
        c = connect.Client("https://connect.example", "12345")
        permissions = Permissions(c._ctx, content_guid=content_guid)

        # invoke
        count = permissions.count()

        # assert response
        assert count == len(fake_permissions)


class TestPermissionsCreate:
    @responses.activate
    def test(self):
        # data
        uid = "94"
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        principal_guid = str(uuid.uuid4())
        principal_type = "user"
        role = "owner"
        fake_permission = {
            **load_mock_dict(f"v1/content/{content_guid}/permissions/{uid}.json"),
            "principal_guid": principal_guid,
            "principal_type": principal_type,
            "role": role,
        }

        # behavior
        responses.post(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions",
            json=fake_permission,
            match=[
                matchers.json_params_matcher(
                    {
                        "principal_guid": principal_guid,
                        "principal_type": principal_type,
                        "role": role,
                    },
                ),
            ],
        )

        # setup
        c = connect.Client("https://connect.example", "12345")
        permissions = Permissions(c._ctx, content_guid=content_guid)

        # invoke
        permission = permissions.create(
            principal_guid=principal_guid,
            principal_type=principal_type,
            role=role,
        )

        # assert
        assert permission == fake_permission

    def test_assertions(self):
        # setup
        principal_guid = "principal_guid"
        content_guid = "content_guid"
        c = Client("https://connect.example/__api__", "12345")
        permissions = Permissions(c._ctx, content_guid=content_guid)
        user = User(c._ctx, guid=principal_guid)
        group = User(c._ctx, guid=principal_guid)

        # behavior
        with pytest.raises(TypeError, match="str"):
            permissions.create(  # pyright: ignore[reportCallIssue]
                "not a user or group",
            )
        with pytest.raises(ValueError):
            permissions.create(  # pyright: ignore[reportCallIssue]
                user,
                principal_guid=principal_guid,
            )
        with pytest.raises(ValueError):
            permissions.create(  # pyright: ignore[reportCallIssue]
                user,
                principal_type="viewer",
            )

    @responses.activate
    def test_user_group(self):
        # data
        content_guid = "CONTENT_GUID"
        user_guid = "USER_GUID"
        group_guid = "GROUP_GUID"

        fake_user = {
            "principal_guid": user_guid,
            "principal_type": "user",
            "role": "viewer",
        }
        fake_group = {
            "principal_guid": group_guid,
            "principal_type": "group",
            "role": "viewer",
        }
        res_user = responses.post(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions",
            json=fake_user,
            match=[matchers.json_params_matcher(fake_user)],
        )
        res_group = responses.post(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions",
            json=fake_group,
            match=[matchers.json_params_matcher(fake_group)],
        )

        # setup
        c = Client("https://connect.example/__api__", "12345")
        permissions = Permissions(c._ctx, content_guid=content_guid)
        user = User(c._ctx, guid=user_guid)
        group = Group(c._ctx, guid=group_guid)

        # invoke
        user_perm = permissions.create(user, role="viewer")
        group_perm = permissions.create(group, role="viewer")

        created_permissions = [user_perm, group_perm]

        # assert
        assert res_user.call_count == 1
        assert res_group.call_count == 1

        for permission in created_permissions:
            assert isinstance(permission, Permission)

        assert created_permissions == [
            Permission(
                c._ctx,
                principal_guid=user_guid,
                principal_type="user",
                role="viewer",
            ),
            Permission(
                c._ctx,
                principal_guid=group_guid,
                principal_type="group",
                role="viewer",
            ),
        ]


class TestPermissionsFind:
    @responses.activate
    def test(self):
        # test data
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fake_permissions = load_mock(f"v1/content/{content_guid}/permissions.json")

        # define api behavior
        responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions",
            json=fake_permissions,
        )

        # setup
        c = connect.Client("https://connect.example", "12345")
        permissions = Permissions(c._ctx, content_guid=content_guid)

        # invoke
        permissions = permissions.find()

        # assert response
        assert permissions == fake_permissions


class TestPermissionsFindOne:
    @responses.activate
    def test(self):
        # test data
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fake_permissions = load_mock_list(f"v1/content/{content_guid}/permissions.json")

        # define api behavior
        responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions",
            json=fake_permissions,
        )

        # setup
        c = connect.Client("https://connect.example", "12345")
        permissions = Permissions(c._ctx, content_guid=content_guid)

        # invoke
        permission = permissions.find_one()

        # assert response
        assert permission == fake_permissions[0]


class TestPermissionsGet:
    @responses.activate
    def test(self):
        # data
        uid = "94"
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fake_permission = load_mock(f"v1/content/{content_guid}/permissions/{uid}.json")

        # behavior
        responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/permissions/{uid}",
            json=fake_permission,
        )

        # setup
        c = connect.Client("https://connect.example", "12345")
        permissions = Permissions(c._ctx, content_guid=content_guid)

        # invoke
        permission = permissions.get(uid)

        # assert
        assert permission == fake_permission


class TestPermissionsDestroy:
    @responses.activate
    def test_destroy(self):
        # data
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        fake_permissions = load_mock_list(f"v1/content/{content_guid}/permissions.json")

        assert fake_permissions[0]["principal_type"] == "user"
        user_permission_id = fake_permissions[0]["id"]
        assert fake_permissions[1]["principal_type"] == "group"
        group_permission_id = fake_permissions[1]["id"]
        assert user_permission_id != group_permission_id

        fake_manual_user_permission = load_mock_dict(
            f"v1/content/{content_guid}/permissions/{user_permission_id}.json"
        )

        fake_user = load_mock_dict("v1/user.json")
        fake_group = load_mock_dict("v1/groups/6f300623-1e0c-48e6-a473-ddf630c0c0c3.json")

        # behavior

        # Used in internal for-loop
        mock_permissions_get = [
            # Used to find all permissions when searching for user and group
            responses.get(
                f"https://connect.example/__api__/v1/content/{content_guid}/permissions",
                json=fake_permissions,
            ),
            # Retrieve permissions object for permission id
            responses.get(
                f"https://connect.example/__api__/v1/content/{content_guid}/permissions/{user_permission_id}",
                json=fake_manual_user_permission,
            ),
        ]
        # permission delete
        mock_permission_deletes = [
            responses.delete(
                f"https://connect.example/__api__/v1/content/{content_guid}/permissions/{user_permission_id}",
            ),
            responses.delete(
                f"https://connect.example/__api__/v1/content/{content_guid}/permissions/{group_permission_id}",
            ),
        ]

        # setup
        c = Client(api_key="12345", url="https://connect.example/")
        ctx = Context(c)
        permissions = Permissions(ctx, content_guid=content_guid)

        # (Doesn't match any permissions, but that's okay)
        user_to_remove = User(ctx, **fake_user)
        group_to_remove = Group(ctx, **fake_group)
        permission_to_remove = Permission(ctx, **fake_manual_user_permission)

        # invoke
        permissions.destroy(permission_to_remove["id"])
        permissions.destroy(permission_to_remove)
        permissions.destroy(user_to_remove)

        permissions.destroy(group_to_remove)

        # Assert values
        assert mock_permissions_get[0].call_count == 2
        assert mock_permissions_get[1].call_count == 1

        # permission_id, user -> permission_id, permission -> permission_id
        assert mock_permission_deletes[0].call_count == 3
        # group -> permission_id
        assert mock_permission_deletes[1].call_count == 1
        # assert mock_permission_deletes[2].call_count == 1

        # Assert bad input value
        with pytest.raises(TypeError):
            permissions.destroy(
                42  # pyright: ignore[reportArgumentType]
            )



"""System resources."""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Literal, overload

from typing_extensions import TypedDict, Unpack

from .context import ContextManager
from .resources import Active

if TYPE_CHECKING:
    from .context import Context
    from .tasks import Task


class System(ContextManager):
    """System information."""

    def __init__(self, ctx: Context, path: str) -> None:
        super().__init__()
        self._ctx: Context = ctx
        # v1/system
        self._path: str = path

    @property
    def caches(self) -> SystemCaches:
        """
        List all system caches.

        Returns
        -------
        SystemCaches
            Helper class for system caches.

        Examples
        --------
        ```python
        from posit.connect import Client

        client = Client()

        caches = client.system.caches.runtime.find()
        ```

        """
        path = self._path + "/caches"
        return SystemCaches(self._ctx, path)


class SystemCaches(ContextManager):
    """System caches."""

    def __init__(self, ctx: Context, path: str) -> None:
        super().__init__()
        self._ctx: Context = ctx
        # v1/system/caches
        self._path: str = path

    @property
    def runtime(self) -> SystemRuntimeCaches:
        """
        System runtime caches.

        Returns
        -------
        SystemRuntimeCaches
            Helper class to manage system runtime caches.

        Examples
        --------
        ```python
        from posit.connect import Client

        client = Client()

        caches = client.system.caches.runtime.find()
        ```
        """
        path = self._path + "/runtime"
        return SystemRuntimeCaches(self._ctx, path)


class SystemRuntimeCache(Active):
    class _Attrs(TypedDict, total=False):
        language: str  # "Python"
        """The runtime language of the cache."""
        version: str  # "3.8.12"
        """The language version of the cache."""
        image_name: str  # "Local"
        """The name of the cache's execution environment."""

    def __init__(
        self,
        ctx: Context,
        path: str,
        /,
        **attributes: Unpack[_Attrs],
    ) -> None:
        super().__init__(ctx, path, **attributes)

    class _DestroyAttrs(TypedDict, total=False):
        dry_run: bool
        """If `True`, the cache will not be destroyed, only the operation will be simulated."""

    @overload
    def destroy(self, *, dry_run: Literal[True]) -> None: ...
    @overload
    def destroy(self, *, dry_run: Literal[False] = False) -> Task: ...
    def destroy(self, **kwargs) -> Task | None:
        """
        Remove a content runtime package cache.

        This action is only available to administrators.

        Parameters
        ----------
        dry_run : bool, optional
            If `True`, the cache will not be destroyed, only the operation will be simulated.

        Returns
        -------
        Task | None
            The task object if the operation was successful. If `dry_run=True`, `None` is returned.

        Examples
        --------
        ```python
        from posit.connect import Client

        client = Client()

        runtime_caches = client.system.caches.runtime.find()
        first_runtime_cache = runtime_caches[0]

        # Remove the cache
        task = first_runtime_cache.destroy(dry_run=False)

        # Wait for the task to finish
        task.wait_for()
        ```
        """
        body = {**self, **kwargs}
        response = self._ctx.client.delete(self._path, json=body)

        task_id = response.json().get("task_id")
        if not task_id:
            return None
        task = self._ctx.client.tasks.get(task_id)
        return task


class SystemRuntimeCaches(ContextManager):
    """
    System runtime caches.

    List all content runtime caches. These include packrat and Python
    environment caches.

    This information is available only to administrators.
    """

    def __init__(self, ctx: Context, path: str) -> None:
        super().__init__()
        self._ctx: Context = ctx
        # v1/system/caches/runtime
        self._path: str = path

    def find(self) -> List[SystemRuntimeCache]:
        """
        List all content runtime caches.

        List all content runtime caches. These include packrat and Python
        environment caches.

        This information is available only to administrators.

        Returns
        -------
        List[SystemRuntimeCache]
            List of all content runtime caches.

        Examples
        --------
        ```python
        from posit.connect import Client

        client = Client()

        runtime_caches = client.system.caches.runtime.find()
        ```
        """
        response = self._ctx.client.get(self._path)
        results = response.json()
        return [SystemRuntimeCache(self._ctx, self._path, **cache) for cache in results["caches"]]

    @overload
    def destroy(
        self,
        /,
        *,
        language: str,
        version: str,
        image_name: str,
        dry_run: Literal[False] = False,
    ) -> Task: ...
    @overload
    def destroy(
        self,
        /,
        *,
        language: str,
        version: str,
        image_name: str,
        dry_run: Literal[True] = True,
    ) -> None: ...

    def destroy(
        self,
        /,
        **kwargs,
    ) -> Task | None:
        """
        Delete a content runtime package cache.

        Delete a content runtime package cache by specifying language, version, and execution
        environment.

        This action is only available to administrators.

        Parameters
        ----------
        language : str
            The runtime language of the cache.
        version : str
            The language version of the cache.
        image_name : str
            The name of the cache's execution environment.
        dry_run : bool, optional
            If `True`, the cache will not be destroyed, only the operation will be simulated.

        Returns
        -------
        Task | None
            The task object if the operation was successful. If `dry_run=True`, `None` is returned.

        Examples
        --------
        ```python
        from posit.connect import Client

        client = Client()

        runtime_caches = client.system.caches.runtime.find()
        first_runtime_cache = runtime_caches[0]

        # Remove the cache
        task = first_runtime_cache.destroy(dry_run=False)

        # Or, remove the cache by specifying the cache's attributes
        task = client.system.caches.runtime.destroy(
            language="Python",
            version="3.12.5",
            image_name="Local",
            dry_run=False,
        )
        ```
        """
        dry_run: bool | None = kwargs.pop("dry_run", None)
        runtime_cache = SystemRuntimeCache(self._ctx, self._path, **kwargs)

        # Only add `dry_run=` argument if it is provided by user
        dry_run_args = {}
        if dry_run is not None:
            dry_run_args["dry_run"] = dry_run
        return runtime_cache.destroy(**dry_run_args)



from posit import connect


class TestTags:
    @classmethod
    def setup_class(cls):
        cls.client = connect.Client()
        cls.contentA = cls.client.content.create(name="Content_A")
        cls.contentB = cls.client.content.create(name="Content_B")
        cls.contentC = cls.client.content.create(name="Content_C")

    @classmethod
    def teardown_class(cls):
        assert len(cls.client.tags.find()) == 0
        cls.contentA.delete()
        cls.contentB.delete()
        cls.contentC.delete()
        assert len(cls.client.content.find()) == 0

    def test_tags_find(self):
        tags = self.client.tags.find()
        assert len(tags) == 0

    def test_tags_create_destroy(self):
        tagA = self.client.tags.create(name="tagA")
        tagB = self.client.tags.create(name="tagB")
        tagC = self.client.tags.create(name="tagC", parent=tagA)

        assert len(self.client.tags.find()) == 3

        tagA.destroy()
        tagB.destroy()
        ## Deleted when tag A is deleted
        # tagC.destroy()

        assert len(self.client.tags.find()) == 0

    def test_tag_descendants(self):
        tagA = self.client.tags.create(name="tagA")
        tagB = self.client.tags.create(name="tagB")
        tagC = self.client.tags.create(name="tagC", parent=tagA)
        tagD = self.client.tags.create(name="tagD", parent=tagC)

        assert tagA.descendant_tags.find() == [tagC, tagD]

        assert len(tagB.descendant_tags.find()) == 0
        assert tagC.descendant_tags.find() == [tagD]

        # cleanup
        tagA.destroy()
        tagB.destroy()
        assert len(self.client.tags.find()) == 0

    def test_tag_children(self):
        tagA = self.client.tags.create(name="tagA_children")
        tagB = self.client.tags.create(name="tagB_children")
        tagC = self.client.tags.create(name="tagC_children", parent=tagA)
        tagD = self.client.tags.create(name="tagD_children", parent=tagC)

        assert tagA.child_tags.find() == [tagC]
        assert tagB.child_tags.find() == []
        assert tagC.child_tags.find() == [tagD]

        # cleanup
        tagA.destroy()
        tagB.destroy()
        assert len(self.client.tags.find()) == 0

    def test_tag_parent(self):
        tagA = self.client.tags.create(name="tagA_parent")
        tagB = self.client.tags.create(name="tagB_parent")
        tagC = self.client.tags.create(name="tagC_parent", parent=tagA)
        tagD = self.client.tags.create(name="tagD_parent", parent=tagC)

        assert tagA.parent_tag is None
        assert tagB.parent_tag is None
        assert tagC.parent_tag == tagA
        assert tagD.parent_tag == tagC

        # cleanup
        tagA.destroy()
        tagB.destroy()
        assert len(self.client.tags.find()) == 0

    def test_content_item_tags(self):
        tagRoot = self.client.tags.create(name="tagRoot_content_item_tags")
        tagA = self.client.tags.create(name="tagA_content_item_tags", parent=tagRoot)
        tagB = self.client.tags.create(name="tagB_content_item_tags", parent=tagRoot)
        tagC = self.client.tags.create(name="tagC_content_item_tags", parent=tagA)
        tagD = self.client.tags.create(name="tagD_content_item_tags", parent=tagC)

        assert len(self.contentA.tags.find()) == 0

        self.contentA.tags.add(tagD)
        self.contentA.tags.add(tagB)

        # tagB, tagD, tagC (parent of tagD), tagA (parent of tagC)
        # tagD + tagC
        # tagRoot is considered a "category" and is "not a tag"
        assert len(self.contentA.tags.find()) == 4

        # Removes tagB
        self.contentA.tags.delete(tagB)
        assert len(self.contentA.tags.find()) == 4 - 1

        # Removes tagC and tagD (parent of tagC)
        self.contentA.tags.delete(tagC)
        assert len(self.contentA.tags.find()) == 4 - 1 - 2

        # cleanup
        tagRoot.destroy()
        assert len(self.client.tags.find()) == 0

    def test_tag_content_items(self):
        tagRoot = self.client.tags.create(name="tagRoot_tag_content_items")
        tagA = self.client.tags.create(name="tagA_tag_content_items", parent=tagRoot)
        tagB = self.client.tags.create(name="tagB_tag_content_items", parent=tagRoot)
        tagC = self.client.tags.create(name="tagC_tag_content_items", parent=tagA)
        tagD = self.client.tags.create(name="tagD_tag_content_items", parent=tagC)

        assert len(tagA.content_items.find()) == 0
        assert len(tagB.content_items.find()) == 0
        assert len(tagC.content_items.find()) == 0
        assert len(tagD.content_items.find()) == 0

        self.contentA.tags.add(tagD)
        self.contentA.tags.add(tagB)

        self.contentB.tags.add(tagA)
        self.contentB.tags.add(tagC)

        self.contentC.tags.add(tagC)

        assert len(self.contentA.tags.find()) == 4
        assert len(self.contentB.tags.find()) == 2
        assert len(self.contentC.tags.find()) == 2

        assert len(tagA.content_items.find()) == 3
        assert len(tagB.content_items.find()) == 1
        assert len(tagC.content_items.find()) == 3

        # Make sure unique content items are found
        content_items = tagA.content_items.find()
        assert len(content_items) == 3

        content_item_guids = {content_item["guid"] for content_item in content_items}
        assert content_item_guids == {
            self.contentA["guid"],
            self.contentB["guid"],
            self.contentC["guid"],
        }

        # Update tag
        tagD.update(name="tagD_updated")
        assert tagD["name"] == "tagD_updated"
        assert self.client.tags.get(tagD["id"])["name"] == "tagD_updated"

        tagD.update(parent=tagB)
        assert tagD["parent_id"] == tagB["id"]
        assert self.client.tags.get(tagD["id"])["parent_id"] == tagB["id"]

        # Cleanup
        self.contentA.tags.delete(tagRoot)
        self.contentB.tags.delete(tagRoot)
        self.contentC.tags.delete(tagRoot)

        assert len(tagA.content_items.find()) == 0
        assert len(tagB.content_items.find()) == 0
        assert len(tagC.content_items.find()) == 0
        assert len(tagD.content_items.find()) == 0

        tagRoot.destroy()
        assert len(self.client.tags.find()) == 0



"""Repository resources."""

from __future__ import annotations

from typing_extensions import (
    Optional,
    Protocol,
    overload,
    runtime_checkable,
)

from ._utils import update_dict_values
from .errors import ClientError
from .resources import Resource, _Resource


# ContentItem Repository uses a PATCH method, not a PUT for updating.
class _ContentItemRepository(_Resource):
    def update(self, **attributes) -> None:
        response = self._ctx.client.patch(self._path, json=attributes)
        result = response.json()

        update_dict_values(self, **result)


@runtime_checkable
class ContentItemRepository(Resource, Protocol):
    """
    Content items GitHub repository information.

    See Also
    --------
    * Get info: https://docs.posit.co/connect/api/#get-/v1/content/-guid-/repository
    * Delete info: https://docs.posit.co/connect/api/#delete-/v1/content/-guid-/repository
    * Update info: https://docs.posit.co/connect/api/#patch-/v1/content/-guid-/repository
    """

    def destroy(self) -> None:
        """
        Delete the content's git repository location.

        See Also
        --------
        * https://docs.posit.co/connect/api/#delete-/v1/content/-guid-/repository
        """
        ...

    def update(
        self,
        *,
        repository: Optional[str] = None,
        branch: str = "main",
        directory: str = ".",
        polling: bool = False,
    ) -> None:
        """Update the content's repository.

        Parameters
        ----------
        repository: str, optional
            URL for the repository. Default is None.
        branch: str, optional
            The tracked Git branch. Default is 'main'.
        directory: str, optional
            Directory containing the content. Default is '.'
        polling: bool, optional
            Indicates that the Git repository is regularly polled. Default is False.

        Returns
        -------
        None

        See Also
        --------
        * https://docs.posit.co/connect/api/#patch-/v1/content/-guid-/repository
        """
        ...


class ContentItemRepositoryMixin:
    @property
    def repository(self: Resource) -> ContentItemRepository | None:
        try:
            path = f"v1/content/{self['guid']}/repository"
            response = self._ctx.client.get(path)
            result = response.json()
            return _ContentItemRepository(
                self._ctx,
                path,
                **result,
            )
        except ClientError:
            return None

    @overload
    def create_repository(
        self: Resource,
        /,
        *,
        repository: Optional[str] = None,
        branch: str = "main",
        directory: str = ".",
        polling: bool = False,
    ) -> ContentItemRepository: ...

    @overload
    def create_repository(self: Resource, /, **attributes) -> ContentItemRepository: ...

    def create_repository(self: Resource, /, **attributes) -> ContentItemRepository:
        """Create repository.

        Parameters
        ----------
        repository : str
            URL for the respository.
        branch : str, optional
            The tracked Git branch. Default is 'main'.
        directory : str, optional
            Directory containing the content. Default is '.'.
        polling : bool, optional
            Indicates that the Git repository is regularly polled. Default is False.

        Returns
        -------
        ContentItemRepository
        """
        path = f"v1/content/{self['guid']}/repository"
        response = self._ctx.client.put(path, json=attributes)
        result = response.json()

        return _ContentItemRepository(
            self._ctx,
            path,
            **result,
        )



"""Group resources."""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, overload

from .paginator import Paginator
from .resources import BaseResource, Resources

if TYPE_CHECKING:
    import requests

    from .context import Context
    from .users import User


class Group(BaseResource):
    def __init__(self, ctx: Context, **kwargs) -> None:
        super().__init__(ctx, **kwargs)
        self._ctx: Context = ctx

    @property
    def members(self) -> GroupMembers:
        """Get the group members.

        Returns
        -------
        GroupMembers
            All the users in the group.

        Examples
        --------
        ```python
        from posit.connect import Client

        client = Client("https://posit.example.com", "API_KEY")

        group = client.groups.get("GROUP_GUID_HERE")
        group_users = group.members.find()

        # Get count of group members
        group_user_count = group.members.count()
        ```

        """
        return GroupMembers(self._ctx, group_guid=self["guid"])

    def delete(self) -> None:
        """Delete the group.

        Examples
        --------
        ```python
        from posit.connect import Client

        client = Client("https://posit.example.com", "API_KEY")

        group = client.groups.get("GROUP_GUID_HERE")

        # Delete the group
        group.delete()
        ```
        """
        self._ctx.client.delete(f"v1/groups/{self['guid']}")


class GroupMembers(Resources):
    def __init__(self, ctx: Context, group_guid: str) -> None:
        super().__init__(ctx)
        self._group_guid = group_guid

    @overload
    def add(self, user: User, /) -> None: ...
    @overload
    def add(self, /, *, user_guid: str) -> None: ...

    def add(self, user: Optional[User] = None, /, *, user_guid: Optional[str] = None) -> None:
        """Add a user to the group.

        Parameters
        ----------
        user : User
            User object to add to the group. Only one of `user=` or `user_guid=` can be provided.
        user_guid : str
            The user GUID.

        Examples
        --------
        ```python
        from posit.connect import Client

        client = Client("https://posit.example.com", "API_KEY")

        group = client.groups.get("GROUP_GUID_HERE")
        user = client.users.get("USER_GUID_HERE")

        # Add a user to the group
        group.members.add(user)

        # Add multiple users to the group
        users = client.users.find()
        for user in users:
            group.members.add(user)

        # Add a user to the group by GUID
        group.members.add(user_guid="USER_GUID_HERE")
        ```

        See Also
        --------
        * https://docs.posit.co/connect/api/#post-/v1/groups/-group_guid-/members
        """
        if user is not None:
            from .users import User

            if user_guid:
                raise ValueError("Only one of `user=` or `user_guid=` should be provided.")
            if not isinstance(user, User):
                raise TypeError(f"`user=` is not a `User` object. Received {user}")

            user_guid = user["guid"]

        if not isinstance(user_guid, str):
            raise TypeError(f"`user_guid=` should be a string. Received {user_guid}")
        if not user_guid:
            raise ValueError("`user_guid=` should not be empty.")

        self._ctx.client.post(
            f"v1/groups/{self._group_guid}/members",
            json={"user_guid": user_guid},
        )

    @overload
    def delete(self, user: User, /) -> None: ...
    @overload
    def delete(self, /, *, user_guid: str) -> None: ...

    def delete(self, user: Optional[User] = None, /, *, user_guid: Optional[str] = None) -> None:
        """Remove a user from the group.

        Parameters
        ----------
        user : User
            User object to add to the group. Only one of `user=` or `user_guid=` can be provided.
        user_guid : str
            The user GUID.

        Examples
        --------
        ```python
        from posit.connect import Client

        client = Client("https://posit.example.com", "API_KEY")

        group = client.groups.get("GROUP_GUID_HERE")

        # Remove a user from the group
        first_user = group.members.find()[0]
        group.members.delete(first_user)

        # Remove multiple users from the group
        group_users = group.members.find()[:2]
        for group_user in group_users:
            group.members.delete(group_user)

        # Remove a user from the group by GUID
        group.members.delete(user_guid="USER_GUID_HERE")
        ```

        See Also
        --------
        * https://docs.posit.co/connect/api/#delete-/v1/groups/-group_guid-/members/-user_guid-
        """
        if user is not None:
            from .users import User

            if user_guid:
                raise ValueError("Only one of `user=` or `user_guid=` should be provided.")
            if not isinstance(user, User):
                raise TypeError(f"`user=` is not a `User` object. Received {user}")

            user_guid = user["guid"]

        if not isinstance(user_guid, str):
            raise TypeError(f"`user_guid=` should be a string. Received {user_guid}")
        if not user_guid:
            raise ValueError("`user_guid=` should not be empty.")

        self._ctx.client.delete(f"v1/groups/{self._group_guid}/members/{user_guid}")

    def find(self) -> list[User]:
        """Find group members.

        Returns
        -------
        list[User]
            All the users in the group.

        Examples
        --------
        ```python
        from posit.connect import Client

        client = Client("https://posit.example.com", "API_KEY")

        group = client.groups.get("GROUP_GUID_HERE")

        # Find all users in the group
        group_users = group.members.find()
        ```

        See Also
        --------
        * https://docs.posit.co/connect/api/#get-/v1/groups/-group_guid-/members
        """
        # Avoid circular import
        from .users import User

        path = f"v1/groups/{self._group_guid}/members"
        paginator = Paginator(self._ctx, path)
        member_dicts = paginator.fetch_results()

        # For each member in the group
        users = [User(self._ctx, **member_dict) for member_dict in member_dicts]
        return users

    def count(self) -> int:
        """Count the number of group members.

        Returns
        -------
        int

        Examples
        --------
        ```python
        from posit.connect import Client

        client = Client("https://posit.example.com", "API_KEY")

        group = client.groups.get("GROUP_GUID_HERE")

        # Get count of group members
        group_user_count = group.members.count()
        ```

        See Also
        --------
        * https://docs.posit.co/connect/api/#get-/v1/groups/-group_guid-/members
        """
        response = self._ctx.client.get(
            f"v1/groups/{self._group_guid}/members",
            params={"page_size": 1},
        )
        result = response.json()
        return result["total"]


class Groups(Resources):
    """Groups resource."""

    @overload
    def create(self, *, name: str, unique_id: str | None) -> Group:
        """Create a group.

        Parameters
        ----------
        name: str
        unique_id: str | None

        Returns
        -------
        Group

        See Also
        --------
        * https://docs.posit.co/connect/api/#post-/v1/groups
        """

    @overload
    def create(self, **kwargs) -> Group:
        """Create a group.

        Returns
        -------
        Group
        """

    def create(self, **kwargs) -> Group:
        """Create a group.

        Parameters
        ----------
        name: str
        unique_id: str | None

        Returns
        -------
        Group
        """
        response = self._ctx.client.post("v1/groups", json=kwargs)
        return Group(self._ctx, **response.json())

    @overload
    def find(
        self,
        *,
        prefix: str = ...,
    ) -> List[Group]: ...

    @overload
    def find(self, **kwargs) -> List[Group]: ...

    def find(self, **kwargs):
        """Find groups.

        Parameters
        ----------
        prefix: str
            Filter by group name prefix. Casing is ignored.

        Returns
        -------
        List[Group]

        See Also
        --------
        * https://docs.posit.co/connect/api/#get-/v1/groups
        """
        path = "v1/groups"
        paginator = Paginator(self._ctx, path, params=kwargs)
        results = paginator.fetch_results()
        return [
            Group(
                self._ctx,
                **result,
            )
            for result in results
        ]

    @overload
    def find_one(
        self,
        *,
        prefix: str = ...,
    ) -> Group | None: ...

    @overload
    def find_one(self, **kwargs) -> Group | None: ...

    def find_one(self, **kwargs) -> Group | None:
        """Find one group.

        Parameters
        ----------
        prefix: str
            Filter by group name prefix. Casing is ignored.

        Returns
        -------
        Group | None

        See Also
        --------
        * https://docs.posit.co/connect/api/#get-/v1/groups
        """
        path = "v1/groups"
        paginator = Paginator(self._ctx, path, params=kwargs)
        pages = paginator.fetch_pages()
        results = (result for page in pages for result in page.results)
        groups = (
            Group(
                self._ctx,
                **result,
            )
            for result in results
        )
        return next(groups, None)

    def get(self, guid: str) -> Group:
        """Get group.

        Parameters
        ----------
        guid : str

        Returns
        -------
        Group

        See Also
        --------
        * https://docs.posit.co/connect/api/#get-/v1/groups
        """
        response = self._ctx.client.get(f"v1/groups/{guid}")
        return Group(
            self._ctx,
            **response.json(),
        )

    def count(self) -> int:
        """Count the number of groups.

        Returns
        -------
        int

        See Also
        --------
        * https://docs.posit.co/connect/api/#get-/v1/groups
        """
        path = "v1/groups"
        response: requests.Response = self._ctx.client.get(path, params={"page_size": 1})
        result: dict = response.json()
        return result["total"]



from pathlib import Path

import pytest
from packaging import version

from posit.connect import Client
from posit.connect.system import SystemRuntimeCache
from posit.connect.tasks import Task

from . import CONNECT_VERSION


@pytest.mark.skipif(
    # Added to the v2023.05.0 milestone
    # https://github.com/rstudio/connect/pull/23148
    CONNECT_VERSION < version.parse("2023.05.0"),
    reason="Cache runtimes not implemented",
)
class TestSystem:
    @classmethod
    def setup_class(cls):
        cls.client = Client()
        assert cls.client.content.count() == 0
        cls.content_item = cls.client.content.create(name="Content_A")

    # Copied from from integration/tests/posit/connect/test_packages.py
    def _deploy_python_bundle(self):
        path = Path("../../../resources/connect/bundles/example-flask-minimal/bundle.tar.gz")
        path = (Path(__file__).parent / path).resolve()
        bundle = self.content_item.bundles.create(str(path))
        task = bundle.deploy()
        task.wait_for()

    def _remove_all_caches(self):
        caches: list[SystemRuntimeCache] = self.client.system.caches.runtime.find()
        for cache in caches:
            assert isinstance(cache, SystemRuntimeCache)
            none_val = cache.destroy(dry_run=True)
            assert none_val is None
            task: Task = cache.destroy()
            assert isinstance(task, Task)
            task.wait_for()
        assert len(self.client.system.caches.runtime.find()) == 0

    @classmethod
    def teardown_class(cls):
        cls.content_item.delete()
        assert cls.client.content.count() == 0

    def test_runtime_caches(self):
        # Get current caches
        caches: list[SystemRuntimeCache] = self.client.system.caches.runtime.find()
        assert isinstance(caches, list)

        # Remove all caches
        self._remove_all_caches()

        # Deploy a new cache
        self._deploy_python_bundle()

        # Check if the cache is deployed
        caches = self.client.system.caches.runtime.find()

        # Barret 2024/12:
        # Caches only showing up in Connect versions >= 2024.05.0
        # I do not know why.
        # Since we are not logic testing Connect, we can confirm our code works given more recent versions of Connect.
        if CONNECT_VERSION >= version.parse("2024.05.0"):
            assert len(caches) > 0

        # Remove all caches
        self._remove_all_caches()



import io
from unittest import mock

import pytest
import responses
from responses import matchers

from posit.connect import Client

from .api import get_path, load_mock


class TestBundleDelete:
    @responses.activate
    def test(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundle_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}",
            json=load_mock(f"v1/content/{content_guid}/bundles/{bundle_id}.json"),
        )

        mock_bundle_delete = responses.delete(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}",
        )

        # setup
        c = Client("https://connect.example", "12345")
        bundle = c.content.get(content_guid).bundles.get(bundle_id)

        # invoke
        bundle.delete()

        # assert
        assert mock_content_get.call_count == 1
        assert mock_bundle_get.call_count == 1
        assert mock_bundle_delete.call_count == 1


class TestBundleDeploy:
    @responses.activate
    def test(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"
        task_id = "jXhOhdm5OOSkGhJw"

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundle_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}",
            json=load_mock(f"v1/content/{content_guid}/bundles/{bundle_id}.json"),
        )

        mock_bundle_deploy = responses.post(
            f"https://connect.example/__api__/v1/content/{content_guid}/deploy",
            match=[matchers.json_params_matcher({"bundle_id": bundle_id})],
            json={"task_id": task_id},
        )

        mock_tasks_get = responses.get(
            f"https://connect.example/__api__/v1/tasks/{task_id}",
            json=load_mock(f"v1/tasks/{task_id}.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")
        bundle = c.content.get(content_guid).bundles.get(bundle_id)

        # invoke
        task = bundle.deploy()

        # assert
        assert task["id"] == task_id
        assert mock_content_get.call_count == 1
        assert mock_bundle_get.call_count == 1
        assert mock_bundle_deploy.call_count == 1
        assert mock_tasks_get.call_count == 1


class TestBundleDownload:
    @mock.patch("builtins.open", new_callable=mock.mock_open)
    @responses.activate
    def test_output_as_str(self, mock_file: mock.MagicMock):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"
        path = get_path(f"v1/content/{content_guid}/bundles/{bundle_id}/download/bundle.tar.gz")

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundle_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}",
            json=load_mock(f"v1/content/{content_guid}/bundles/{bundle_id}.json"),
        )

        mock_bundle_download = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}/download",
            body=path.read_bytes(),
        )

        # setup
        c = Client("https://connect.example", "12345")
        bundle = c.content.get(content_guid).bundles.get(bundle_id)

        # invoke
        bundle.download("pathname")

        # assert
        assert mock_content_get.call_count == 1
        assert mock_bundle_get.call_count == 1
        assert mock_bundle_download.call_count == 1
        mock_file.assert_called_once_with("pathname", "wb")

    @responses.activate
    def test_output_as_io(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"
        path = get_path(f"v1/content/{content_guid}/bundles/{bundle_id}/download/bundle.tar.gz")

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundle_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}",
            json=load_mock(f"v1/content/{content_guid}/bundles/{bundle_id}.json"),
        )

        mock_bundle_download = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}/download",
            body=path.read_bytes(),
        )

        # setup
        c = Client("https://connect.example", "12345")
        bundle = c.content.get(content_guid).bundles.get(bundle_id)

        # invoke
        file = io.BytesIO()
        buffer = io.BufferedWriter(
            file  # pyright: ignore[reportArgumentType]
        )
        bundle.download(buffer)
        buffer.seek(0)

        # assert
        assert mock_content_get.call_count == 1
        assert mock_bundle_get.call_count == 1
        assert mock_bundle_download.call_count == 1
        assert file.read() == path.read_bytes()

    @responses.activate
    def test_invalid_arguments(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"
        path = get_path(f"v1/content/{content_guid}/bundles/{bundle_id}/download/bundle.tar.gz")

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundle_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}",
            json=load_mock(f"v1/content/{content_guid}/bundles/{bundle_id}.json"),
        )

        mock_bundle_download = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}/download",
            body=path.read_bytes(),
        )

        # setup
        c = Client("https://connect.example", "12345")
        bundle = c.content.get(content_guid).bundles.get(bundle_id)

        # invoke
        with pytest.raises(TypeError):
            bundle.download(
                None  # pyright: ignore[reportArgumentType]
            )

        # assert
        assert mock_content_get.call_count == 1
        assert mock_bundle_get.call_count == 1


class TestBundlesCreate:
    @responses.activate
    def test(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"
        pathname = get_path(
            f"v1/content/{content_guid}/bundles/{bundle_id}/download/bundle.tar.gz",
        )

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundle_post = responses.post(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles",
            json=load_mock(f"v1/content/{content_guid}/bundles/{bundle_id}.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")
        content = c.content.get(content_guid)

        # invoke
        data = pathname.read_bytes()
        bundle = content.bundles.create(data)

        # # assert
        assert bundle["id"] == "101"
        assert mock_content_get.call_count == 1
        assert mock_bundle_post.call_count == 1

    @responses.activate
    def test_kwargs_pathname(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"
        pathname = get_path(
            f"v1/content/{content_guid}/bundles/{bundle_id}/download/bundle.tar.gz",
        )

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundle_post = responses.post(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles",
            json=load_mock(f"v1/content/{content_guid}/bundles/{bundle_id}.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")
        content = c.content.get(content_guid)

        # invoke
        pathname = str(pathname.absolute())
        bundle = content.bundles.create(pathname)

        # # assert
        assert bundle["id"] == "101"
        assert mock_content_get.call_count == 1
        assert mock_bundle_post.call_count == 1

    @responses.activate
    def test_invalid_arguments(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")
        content = c.content.get(content_guid)

        # invoke
        with pytest.raises(TypeError):
            content.bundles.create(
                None  # pyright: ignore[reportArgumentType]
            )


class TestBundlesFind:
    @responses.activate
    def test(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundles_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles",
            json=load_mock(f"v1/content/{content_guid}/bundles.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")

        # invoke
        bundles = c.content.get(content_guid).bundles.find()

        # assert
        assert mock_content_get.call_count == 1
        assert mock_bundles_get.call_count == 1
        assert len(bundles) == 1
        assert bundles[0]["id"] == "101"


class TestBundlesFindOne:
    @responses.activate
    def test(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundles_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles",
            json=load_mock(f"v1/content/{content_guid}/bundles.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")

        # invoke
        bundle = c.content.get(content_guid).bundles.find_one()

        # assert
        assert mock_content_get.call_count == 1
        assert mock_bundles_get.call_count == 1
        assert bundle
        assert bundle["id"] == "101"


class TestBundlesGet:
    @responses.activate
    def test(self):
        content_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        bundle_id = "101"

        # behavior
        mock_content_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}",
            json=load_mock(f"v1/content/{content_guid}.json"),
        )

        mock_bundle_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_guid}/bundles/{bundle_id}",
            json=load_mock(f"v1/content/{content_guid}/bundles/{bundle_id}.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")

        # invoke
        bundle = c.content.get(content_guid).bundles.get(bundle_id)

        # assert
        assert mock_content_get.call_count == 1
        assert mock_bundle_get.call_count == 1
        assert bundle["id"] == bundle_id



# -*- coding: utf-8 -*-
# mypy: ignore-errors
import os

from databricks import sql
from databricks.sdk.core import Config, databricks_cli
from flask import Flask, request

from posit.connect.external.databricks import PositCredentialsStrategy

DATABRICKS_HOST = os.getenv("DATABRICKS_HOST")
DATABRICKS_HOST_URL = f"https://{DATABRICKS_HOST}"
SQL_HTTP_PATH = os.getenv("DATABRICKS_PATH")

rows = None
app = Flask(__name__)


@app.route("/")
def usage():
    return "<p>Try: <pre>GET /fares<pre></p>"


@app.route("/fares")
def get_fares():
    """
    Flask example API that returns the first few rows from
    a table hosted in Databricks.
    """
    global rows

    session_token = request.headers.get("Posit-Connect-User-Session-Token")
    posit_strategy = PositCredentialsStrategy(
        local_strategy=databricks_cli,
        user_session_token=session_token,
    )
    cfg = Config(
        host=DATABRICKS_HOST_URL,
        # uses Posit's custom credential_strategy if running on Connect,
        # otherwise falls back to the strategy defined by local_strategy
        credentials_strategy=posit_strategy,
    )

    if rows is None:
        query = "SELECT * FROM samples.nyctaxi.trips LIMIT 10;"

        with sql.connect(
            server_hostname=DATABRICKS_HOST,
            http_path=SQL_HTTP_PATH,
            # https://github.com/databricks/databricks-sql-python/issues/148#issuecomment-2271561365
            credentials_provider=posit_strategy.sql_credentials_provider(cfg),
        ) as connection:
            with connection.cursor() as cursor:
                cursor.execute(query)
                rows = cursor.fetchall()

    return [row.asDict() for row in rows]


if __name__ == "__main__":
    app.run(debug=True)



from unittest import mock
from unittest.mock import Mock

import pytest
import responses

from posit.connect.client import Client
from posit.connect.context import Context
from posit.connect.groups import Group
from posit.connect.users import User

from .api import load_mock_dict

session = Mock()
url = Mock()


class TestGroupAttributes:
    @classmethod
    def setup_class(cls):
        guid = "6f300623-1e0c-48e6-a473-ddf630c0c0c3"
        fake_item = load_mock_dict(f"v1/groups/{guid}.json")
        cls.item = Group(mock.Mock(), **fake_item)

    def test_guid(self):
        assert self.item["guid"] == "6f300623-1e0c-48e6-a473-ddf630c0c0c3"

    def test_name(self):
        assert self.item["name"] == "Friends"

    def test_owner_guid(self):
        assert self.item["owner_guid"] == "20a79ce3-6e87-4522-9faf-be24228800a4"


class TestGroupMembers:
    @classmethod
    def setup_class(cls):
        cls.client = Client("https://connect.example", "12345")
        guid = "6f300623-1e0c-48e6-a473-ddf630c0c0c3"
        fake_item = load_mock_dict(f"v1/groups/{guid}.json")
        ctx = Context(cls.client)
        cls.group = Group(ctx, **fake_item)

    @responses.activate
    def test_members_count(self):
        responses.get(
            f"https://connect.example/__api__/v1/groups/{self.group['guid']}/members",
            json=load_mock_dict(f"v1/groups/{self.group['guid']}/members.json"),
        )
        group_members = self.group.members

        assert group_members.count() == 2

    @responses.activate
    def test_members_find(self):
        responses.get(
            f"https://connect.example/__api__/v1/groups/{self.group['guid']}/members",
            json=load_mock_dict(f"v1/groups/{self.group['guid']}/members.json"),
        )

        group_users = self.group.members.find()
        assert len(group_users) == 2
        for user in group_users:
            assert isinstance(user, User)

    @responses.activate
    def test_members_add(self):
        user_guid = "user-guid"
        responses.post(
            f"https://connect.example/__api__/v1/groups/{self.group['guid']}/members",
            json=[],  # No need to return anything
        )

        user = User(self.client._ctx, guid=user_guid)
        self.group.members.add(user)
        self.group.members.add(user_guid=user["guid"])

        with pytest.raises(TypeError):
            self.group.members.add(
                "not-a-user",  # pyright: ignore[reportArgumentType]
            )
        with pytest.raises(TypeError):
            self.group.members.add(group_guid=42)  # pyright: ignore[reportCallIssue]
        with pytest.raises(ValueError):
            self.group.members.add(user, user_guid=user["guid"])  # pyright: ignore[reportCallIssue]
        with pytest.raises(ValueError):
            self.group.members.add(user_guid="")

    @responses.activate
    def test_members_delete(self):
        user_guid = "user-guid"
        responses.get(
            f"https://connect.example/__api__/v1/groups/{self.group['guid']}",
            json=dict(self.group),
        )
        responses.delete(
            f"https://connect.example/__api__/v1/groups/{self.group['guid']}/members/{user_guid}",
            json=[],  # No need to return anything
        )

        user = User(self.client._ctx, guid=user_guid)

        self.group.members.delete(user)
        self.group.members.delete(user_guid=user["guid"])

        with pytest.raises(TypeError):
            self.group.members.delete(
                "not-a-user",  # pyright: ignore[reportArgumentType]
            )
        with pytest.raises(TypeError):
            self.group.members.delete(group_guid=42)  # pyright: ignore[reportCallIssue]
        with pytest.raises(ValueError):
            self.group.members.delete(user, user_guid=user["guid"])  # pyright: ignore[reportCallIssue]

        with pytest.raises(ValueError):
            self.group.members.delete(user_guid="")



"""Snowflake SDK integration.

Snowflake SDK credentials implementations which support interacting with Posit OAuth integrations on Connect.

Notes
-----
The APIs in this module are provided as a convenience and are subject to breaking changes.
"""

from typing import Optional

from ..client import Client
from .external import is_local


class PositAuthenticator:
    """
    Authenticator for Snowflake SDK which supports Posit OAuth integrations on Connect.

    Examples
    --------
    ```python
    import os

    import pandas as pd
    import snowflake.connector
    import streamlit as st

    from posit.connect.external.snowflake import PositAuthenticator

    ACCOUNT = os.getenv("SNOWFLAKE_ACCOUNT")
    WAREHOUSE = os.getenv("SNOWFLAKE_WAREHOUSE")

    # USER is only required when running the example locally with external browser auth
    USER = os.getenv("SNOWFLAKE_USER")

    # https://docs.snowflake.com/en/user-guide/sample-data-using
    DATABASE = os.getenv("SNOWFLAKE_DATABASE", "snowflake_sample_data")
    SCHEMA = os.getenv("SNOWFLAKE_SCHEMA", "tpch_sf1")
    TABLE = os.getenv("SNOWFLAKE_TABLE", "lineitem")

    session_token = st.context.headers.get("Posit-Connect-User-Session-Token")
    auth = PositAuthenticator(
        local_authenticator="EXTERNALBROWSER", user_session_token=session_token
    )

    con = snowflake.connector.connect(
        user=USER,
        account=ACCOUNT,
        warehouse=WAREHOUSE,
        database=DATABASE,
        schema=SCHEMA,
        authenticator=auth.authenticator,
        token=auth.token,
    )

    snowflake_user = con.cursor().execute("SELECT CURRENT_USER()").fetchone()
    st.write(f"Hello, {snowflake_user[0]}!")

    with st.spinner("Loading data from Snowflake..."):
        df = pd.read_sql_query(f"SELECT * FROM {TABLE} LIMIT 10", con)

    st.dataframe(df)
    ```
    """

    def __init__(
        self,
        local_authenticator: Optional[str] = None,
        client: Optional[Client] = None,
        user_session_token: Optional[str] = None,
    ):
        self._local_authenticator = local_authenticator
        self._client = client
        self._user_session_token = user_session_token

    @property
    def authenticator(self) -> Optional[str]:
        if is_local():
            return self._local_authenticator
        return "oauth"

    @property
    def token(self) -> Optional[str]:
        if is_local():
            return None

        # If the user-session-token wasn't provided and we're running on Connect then we raise an exception.
        # user_session_token is required to impersonate the viewer.
        if self._user_session_token is None:
            raise ValueError("The user-session-token is required for viewer authentication.")

        if self._client is None:
            self._client = Client()

        credentials = self._client.oauth.get_credentials(self._user_session_token)
        return credentials.get("access_token")



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from posit import connect

if TYPE_CHECKING:
    from posit.connect.content import ContentItem
    from posit.connect.permissions import Permission


class TestContentPermissions:
    content: ContentItem

    @classmethod
    def setup_class(cls):
        cls.client = connect.Client()
        cls.content = cls.client.content.create(name="example")

        cls.user_aron = cls.client.users.create(
            username="permission_aron",
            email="permission_aron@example.com",
            password="permission_s3cur3p@ssword",
        )
        cls.user_bill = cls.client.users.create(
            username="permission_bill",
            email="permission_bill@example.com",
            password="permission_s3cur3p@ssword",
        )

        cls.group_friends = cls.client.groups.create(name="Friends")

    @classmethod
    def teardown_class(cls):
        cls.content.delete()
        assert cls.client.content.count() == 0

        cls.group_friends.delete()
        assert cls.client.groups.count() == 0

    def test_permissions_add_destroy(self):
        assert self.client.groups.count() == 1
        assert self.client.users.count() == 3
        assert self.content.permissions.find() == []

        # Add permissions
        self.content.permissions.create(
            principal_guid=self.user_aron["guid"],
            principal_type="user",
            role="viewer",
        )
        self.content.permissions.create(
            principal_guid=self.group_friends["guid"],
            principal_type="group",
            role="owner",
        )

        def assert_permissions_match_guids(permissions: list[Permission], objs_with_guid):
            for permission, obj_with_guid in zip(permissions, objs_with_guid):
                assert permission["principal_guid"] == obj_with_guid["guid"]

        # Prove they have been added
        assert_permissions_match_guids(
            self.content.permissions.find(),
            [self.user_aron, self.group_friends],
        )

        # Remove permissions (and from some that isn't an owner)
        self.content.permissions.destroy(self.user_aron)
        with pytest.raises(ValueError):
            self.content.permissions.destroy(self.user_bill)

        # Prove they have been removed
        assert_permissions_match_guids(
            self.content.permissions.find(),
            [self.group_friends],
        )

        # Remove the last permission
        self.content.permissions.destroy(self.group_friends)

        # Prove they have been removed
        assert self.content.permissions.find() == []



from .context import Context
from .users import User


def get(ctx: Context) -> User:
    """
    Gets the current user.

    Args:
        config (Config): The configuration object containing the URL.
        session (requests.Session): The session object used for making HTTP requests.

    Returns
    -------
        User: The current user.
    """
    path = "v1/user"
    response = ctx.client.get(path)
    return User(ctx, **response.json())



"""Vanity URL resources."""

from typing import Callable, List, Optional

from typing_extensions import NotRequired, Required, TypedDict, Unpack

from .context import Context
from .errors import ClientError
from .resources import BaseResource, Resources


class Vanity(BaseResource):
    """A vanity resource.

    Vanities maintain custom URL paths assigned to content.

    Warnings
    --------
    Vanity paths may only contain alphanumeric characters, hyphens, underscores, and slashes.

    Vanities cannot have children. For example, if the vanity path "/finance/" exists, the vanity path "/finance/budget/" cannot. But, if "/finance" does not exist, both "/finance/budget/" and "/finance/report" are allowed.

    The following vanities are reserved by Connect:
    - `/__`
    - `/favicon.ico`
    - `/connect`
    - `/apps`
    - `/users`
    - `/groups`
    - `/setpassword`
    - `/user-completion`
    - `/confirm`
    - `/recent`
    - `/reports`
    - `/plots`
    - `/unpublished`
    - `/settings`
    - `/metrics`
    - `/tokens`
    - `/help`
    - `/login`
    - `/welcome`
    - `/register`
    - `/resetpassword`
    - `/content`
    """

    AfterDestroyCallback = Callable[[], None]

    class VanityAttributes(TypedDict):
        """Vanity attributes."""

        path: Required[str]
        content_guid: Required[str]
        created_time: Required[str]

    def __init__(
        self,
        /,
        ctx: Context,
        *,
        after_destroy: Optional[AfterDestroyCallback] = None,
        **kwargs: Unpack[VanityAttributes],
    ):
        """Initialize a Vanity.

        Parameters
        ----------
        ctx : Context
        after_destroy : AfterDestroyCallback, optional
            Called after the Vanity is successfully destroyed, by default None
        """
        super().__init__(ctx, **kwargs)
        self._after_destroy = after_destroy
        self._content_guid = kwargs["content_guid"]

    def destroy(self) -> None:
        """Destroy the vanity.

        Raises
        ------
        ValueError
            If the foreign unique identifier is missing or its value is `None`.

        Warnings
        --------
        This operation is irreversible.

        Note
        ----
        This action requires administrator privileges.
        """
        path = f"v1/content/{self._content_guid}/vanity"
        self._ctx.client.delete(path)

        if self._after_destroy:
            self._after_destroy()


class Vanities(Resources):
    """Manages a collection of vanities."""

    def all(self) -> List[Vanity]:
        """Retrieve all vanities.

        Returns
        -------
        List[Vanity]

        Notes
        -----
        This action requires administrator privileges.
        """
        path = "v1/vanities"
        response = self._ctx.client.get(path)
        results = response.json()
        return [Vanity(self._ctx, **result) for result in results]


class VanityMixin(BaseResource):
    """Mixin class to add a vanity attribute to a resource."""

    class HasGuid(TypedDict):
        """Has a guid."""

        guid: Required[str]

    def __init__(self, ctx: Context, **kwargs: Unpack[HasGuid]):
        super().__init__(ctx, **kwargs)
        self._content_guid = kwargs["guid"]
        self._vanity: Optional[Vanity] = None

    @property
    def vanity(self) -> Optional[str]:
        """Get the vanity."""
        if self._vanity:
            return self._vanity["path"]

        try:
            self._vanity = self.find_vanity()
            self._vanity._after_destroy = self.reset_vanity
            return self._vanity["path"]
        except ClientError as e:
            if e.http_status == 404:
                return None
            raise e

    @vanity.setter
    def vanity(self, value: str) -> None:
        """Set the vanity.

        Parameters
        ----------
        value : str
            The vanity path.

        Note
        ----
        This action requires owner or administrator privileges.

        See Also
        --------
        create_vanity
        """
        self._vanity = self.create_vanity(path=value)
        self._vanity._after_destroy = self.reset_vanity

    @vanity.deleter
    def vanity(self) -> None:
        """Destroy the vanity.

        Warnings
        --------
        This operation is irreversible.

        Note
        ----
        This action requires owner or administrator privileges.

        See Also
        --------
        reset_vanity
        """
        # Populate value
        self.vanity  # noqa: B018

        if self._vanity:
            self._vanity.destroy()
        self.reset_vanity()

    def reset_vanity(self) -> None:
        """Unload the cached vanity.

        Forces the next access, if any, to query the vanity from the Connect server.
        """
        self._vanity = None

    class CreateVanityRequest(TypedDict, total=False):
        """A request schema for creating a vanity."""

        path: Required[str]
        """The vanity path (e.g., 'my-dashboard')"""

        force: NotRequired[bool]
        """Whether to force creation of the vanity"""

    def create_vanity(self, **kwargs: Unpack[CreateVanityRequest]) -> Vanity:
        """Create a vanity.

        Parameters
        ----------
        path : str, required
            The path for the vanity.
        force : bool, not required
            Whether to force the creation of the vanity. When True, any other vanity with the same path will be deleted.

        Warnings
        --------
        If setting force=True, the destroy operation performed on the other vanity is irreversible.
        """
        path = f"v1/content/{self._content_guid}/vanity"
        response = self._ctx.client.put(path, json=kwargs)
        result = response.json()
        return Vanity(self._ctx, **result)

    def find_vanity(self) -> Vanity:
        """Find the vanity.

        Returns
        -------
        Vanity
        """
        path = f"v1/content/{self._content_guid}/vanity"
        response = self._ctx.client.get(path)
        result = response.json()
        return Vanity(self._ctx, **result)



"""Job resources."""

from __future__ import annotations

from typing import (
    Iterable,
    Literal,
    Protocol,
    runtime_checkable,
)

from .resources import Resource, ResourceSequence

JobTag = Literal[
    "unknown",
    "build_report",
    "build_site",
    "build_jupyter",
    "packrat_restore",
    "python_restore",
    "configure_report",
    "run_app",
    "run_api",
    "run_tensorflow",
    "run_python_api",
    "run_dash_app",
    "run_streamlit",
    "run_bokeh_app",
    "run_fastapi_app",
    "run_pyshiny_app",
    "render_shiny",
    "run_voila_app",
    "testing",
    "git",
    "val_py_ext_pkg",
    "val_r_ext_pkg",
    "val_r_install",
]


StatusCode = Literal[0, 1, 2]


class Job(Resource, Protocol):
    def destroy(self) -> None:
        """Destroy the job.

        Warnings
        --------
        This operation is irreversible.

        Note
        ----
        This action requires administrator, owner, or collaborator privileges.
        """


@runtime_checkable
class Jobs(ResourceSequence[Job], Protocol):
    def fetch(self) -> Iterable[Job]:
        """Fetch all jobs.

        Fetches all jobs from Connect.

        Returns
        -------
        List[Job]
        """
        ...

    def find(self, key: str, /) -> Job:
        """
        Find a Job by its key.

        Fetches the Job from Connect by it's key.

        Parameters
        ----------
        key : str
            The unique identifier of the Job.

        Returns
        -------
        Jobs
        """
        ...

    def find_by(
        self,
        *,
        # Identifiers
        id: str = ...,  # noqa: A002
        ppid: str | None = ...,
        pid: str = ...,
        key: str = ...,
        remote_id: str | None = ...,
        app_id: str = ...,
        variant_id: str = ...,
        bundle_id: str = ...,
        # Timestamps
        start_time: str = ...,
        end_time: str | None = ...,
        last_heartbeat_time: str = ...,
        queued_time: str | None = ...,
        # Status and Exit Information
        status: StatusCode = ...,
        exit_code: int | None = ...,
        # Environment Information
        hostname: str = ...,
        cluster: str | None = ...,
        image: str | None = ...,
        run_as: str = ...,
        queue_name: str | None = ...,
        tag: JobTag = ...,
    ) -> Job | None:
        """Find the first record matching the specified conditions.

        There is no implied ordering, so if order matters, you should specify it yourself.

        id : str, not required
            A unique identifier for the job.
        ppid : Optional[str], not required
            Identifier of the parent process.
        pid : str, not required
            Identifier of the process running the job.
        key : str, not required
            A unique key to identify this job.
        remote_id : Optional[str], not required
            Identifier for off-host execution configurations.
        app_id : str, not required
            Identifier of the parent content associated with the job.
        variant_id : str, not required
            Identifier of the variant responsible for the job.
        bundle_id : str, not required
            Identifier of the content bundle linked to the job.
        start_time : str, not required
            RFC3339 timestamp indicating when the job started.
        end_time : Optional[str], not required
            RFC3339 timestamp indicating when the job finished.
        last_heartbeat_time : str, not required
            RFC3339 timestamp of the last recorded activity for the job.
        queued_time : Optional[str], not required
            RFC3339 timestamp when the job was added to the queue.
        status : int, not required
            Current status. Options are 0 (Active), 1 (Finished), and 2 (Finalized)
        exit_code : Optional[int], not required
            The job's exit code, available after completion.
        hostname : str, not required
            Name of the node processing the job.
        cluster : Optional[str], not required
            Location where the job runs, either 'Local' or the cluster name.
        image : Optional[str], not required
            Location of the content in clustered environments.
        run_as : str, not required
            UNIX user responsible for executing the job.
        queue_name : Optional[str], not required
            Name of the queue processing the job, relevant for scheduled reports.
        tag : JobTag, not required
            A tag categorizing the job type. Options are build_jupyter, build_report, build_site, configure_report, git, packrat_restore, python_restore, render_shiny, run_api, run_app, run_bokeh_app, run_dash_app, run_fastapi_app, run_pyshiny_app, run_python_api, run_streamlit, run_tensorflow, run_voila_app, testing, unknown, val_py_ext_pkg, val_r_ext_pkg, and val_r_install.

        Returns
        -------
        Job | None

        Note
        ----
        This action requires administrator, owner, or collaborator privileges.
        """
        ...



from unittest.mock import Mock

import pytest
import requests
import responses
from responses import matchers

from posit.connect.client import Client
from posit.connect.groups import Group
from posit.connect.users import User

from .api import load_mock, load_mock_dict

session = Mock()
url = Mock()


class TestUserContent:
    """Check behavior of content attribute."""

    @responses.activate
    def test_find(self):
        """Check GET /v1/content call includes owner_guid query parameter."""
        # behavior
        mock_get_user = responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )

        mock_get_content = responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content?owner_guid=20a79ce3-6e87-4522-9faf-be24228800a4.json"),
            match=[
                matchers.query_param_matcher(
                    {"owner_guid": "20a79ce3-6e87-4522-9faf-be24228800a4"},
                    strict_match=False,
                ),
            ],
        )

        # setup
        c = Client(api_key="12345", url="https://connect.example/")
        user = c.users.get("20a79ce3-6e87-4522-9faf-be24228800a4")

        # invoke
        content = user.content.find()

        # assert
        assert mock_get_user.call_count == 1
        assert mock_get_content.call_count == 1
        assert len(content) == 1


class TestUserLock:
    @responses.activate
    def test_lock(self):
        responses.get(
            "https://connect.example/__api__/v1/users/a1b2c3d4-e5f6-g7h8-i9j0-k1l2m3n4o5p6",
            json=load_mock("v1/users/a1b2c3d4-e5f6-g7h8-i9j0-k1l2m3n4o5p6.json"),
        )
        c = Client(api_key="12345", url="https://connect.example/")
        user = c.users.get("a1b2c3d4-e5f6-g7h8-i9j0-k1l2m3n4o5p6")
        assert user["guid"] == "a1b2c3d4-e5f6-g7h8-i9j0-k1l2m3n4o5p6"

        responses.get(
            "https://connect.example/__api__/v1/user",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )
        responses.post(
            "https://connect.example/__api__/v1/users/a1b2c3d4-e5f6-g7h8-i9j0-k1l2m3n4o5p6/lock",
            match=[responses.matchers.json_params_matcher({"locked": True})],
        )
        user.lock()
        assert user["locked"]

    @responses.activate
    def test_lock_self_true(self):
        responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )
        c = Client(api_key="12345", url="https://connect.example/")
        user = c.users.get("20a79ce3-6e87-4522-9faf-be24228800a4")
        assert user["guid"] == "20a79ce3-6e87-4522-9faf-be24228800a4"

        responses.get(
            "https://connect.example/__api__/v1/user",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )
        responses.post(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4/lock",
            match=[responses.matchers.json_params_matcher({"locked": True})],
        )
        user.lock(force=True)
        assert user["locked"]

    @responses.activate
    def test_lock_self_false(self):
        responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )
        c = Client(api_key="12345", url="https://connect.example/")
        user = c.users.get("20a79ce3-6e87-4522-9faf-be24228800a4")
        assert user["guid"] == "20a79ce3-6e87-4522-9faf-be24228800a4"

        responses.get(
            "https://connect.example/__api__/v1/user",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )
        responses.post(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4/lock",
            match=[responses.matchers.json_params_matcher({"locked": True})],
        )
        with pytest.raises(RuntimeError):
            user.lock(force=False)
        assert not user["locked"]


class TestUserUnlock:
    @responses.activate
    def test_unlock(self):
        responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )
        c = Client(api_key="12345", url="https://connect.example/")
        user = c.users.get("20a79ce3-6e87-4522-9faf-be24228800a4")
        assert user["guid"] == "20a79ce3-6e87-4522-9faf-be24228800a4"

        responses.post(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4/lock",
            match=[responses.matchers.json_params_matcher({"locked": False})],
        )
        user.unlock()
        assert not user["locked"]


class TestUserGroups:
    @responses.activate
    def test_groups(self):
        # Get user
        carlos_guid = "20a79ce3-6e87-4522-9faf-be24228800a4"
        carlos_empty_group_guid = "empty-group-guid"
        responses.get(
            f"https://connect.example/__api__/v1/users/{carlos_guid}",
            json=load_mock_dict(f"v1/users/{carlos_guid}.json"),
        )
        responses.get(
            f"https://connect.example/__api__/v1/groups/{carlos_empty_group_guid}/members?page_number=1&page_size=500",
            json=load_mock_dict(f"v1/groups/{carlos_empty_group_guid}/members.json"),
        )
        responses.get(
            "https://connect.example/__api__/v1/users",
            json=load_mock_dict("v1/users?page_number=1&page_size=500.jsonc"),
        )
        # Get groups
        responses.get(
            "https://connect.example/__api__/v1/groups",
            json=load_mock_dict("v1/groups.json"),
        )
        # Get group members
        group_guid = "6f300623-1e0c-48e6-a473-ddf630c0c0c3"
        responses.get(
            f"https://connect.example/__api__/v1/groups/{group_guid}/members",
            json=load_mock_dict(f"v1/groups/{group_guid}/members.json"),
        )

        client = Client("https://connect.example/", "12345")

        carlos = client.users.get(carlos_guid)
        no_groups = carlos.groups.find()
        assert len(no_groups) == 0

        user = client.users.find()[1]
        assert user["guid"] == "87c12c08-11cd-4de1-8da3-12a7579c4998"

        groups = user.groups.find()
        assert isinstance(groups, list)
        assert len(groups) == 1
        group = groups[0]
        assert isinstance(group, Group)
        assert group["name"] == "Friends"

    @responses.activate
    def test_groups_add(self):
        group_guid = "new-group-guid"
        user_guid = "user-guid"
        responses.get(
            f"https://connect.example/__api__/v1/groups/{group_guid}",
            json={"guid": group_guid},
        )
        responses.post(
            f"https://connect.example/__api__/v1/groups/{group_guid}/members",
            json=[],
        )

        client = Client("https://connect.example/", "12345")

        user = User(client._ctx, guid=user_guid)
        new_group = Group(client._ctx, guid=group_guid)
        # Add via Group
        user.groups.add(new_group)
        # Add via guid
        user.groups.add(new_group["guid"])

        with pytest.raises(ValueError):
            user.groups.add("")

    @responses.activate
    def test_groups_delete(self):
        group_guid = "new-group-guid"
        user_guid = "user-guid"
        responses.get(
            f"https://connect.example/__api__/v1/groups/{group_guid}",
            json={"guid": group_guid},
        )
        responses.delete(
            f"https://connect.example/__api__/v1/groups/{group_guid}/members/{user_guid}",
            json=[],
        )

        client = Client("https://connect.example/", "12345")

        user = User(client._ctx, guid=user_guid)
        group = Group(client._ctx, guid=group_guid)

        # Delete via Group
        user.groups.delete(group)
        # Delete via guid
        user.groups.delete(group["guid"])

        with pytest.raises(ValueError):
            user.groups.delete("")


class TestUsers:
    @responses.activate
    def test_users_get(self):
        responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )

        con = Client(api_key="12345", url="https://connect.example/")
        carlos = con.users.get("20a79ce3-6e87-4522-9faf-be24228800a4")
        assert carlos["username"] == "carlos12"
        assert carlos["first_name"] == "Carlos"
        assert carlos["created_time"] == "2019-09-09T15:24:32Z"

    @responses.activate
    def test_users_get_extra_fields(self):
        responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json={
                "guid": "20a79ce3-6e87-4522-9faf-be24228800a4",
                "username": "carlos12",
                "some_new_field": "some_new_value",
            },
        )

        con = Client(api_key="12345", url="https://connect.example/")
        carlos = con.users.get("20a79ce3-6e87-4522-9faf-be24228800a4")
        assert carlos["username"] == "carlos12"
        assert carlos["some_new_field"] == "some_new_value"

    @responses.activate
    def test_user_update(self):
        responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )
        patch_request = responses.put(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            match=[responses.matchers.json_params_matcher({"first_name": "Carlitos"})],
            json={"first_name": "Carlitos"},
        )

        con = Client(api_key="12345", url="https://connect.example/")
        carlos = con.users.get("20a79ce3-6e87-4522-9faf-be24228800a4")

        assert patch_request.call_count == 0
        assert carlos["first_name"] == "Carlos"

        carlos.update(first_name="Carlitos")

        assert patch_request.call_count == 1
        assert carlos["first_name"] == "Carlitos"

    @responses.activate
    def test_user_update_server_error(self):
        responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )
        responses.put(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            status=500,
        )

        con = Client(api_key="12345", url="https://connect.example/")
        carlos = con.users.get("20a79ce3-6e87-4522-9faf-be24228800a4")
        with pytest.raises(requests.HTTPError, match="500 Server Error"):
            carlos.update(first_name="Carlitos")

    @responses.activate
    def test_count(self):
        responses.get(
            "https://connect.example/__api__/v1/users",
            json=load_mock("v1/users?page_number=1&page_size=500.jsonc"),
            match=[responses.matchers.query_param_matcher({"page_size": 1})],
        )
        con = Client(api_key="12345", url="https://connect.example/")
        count = con.users.count()
        assert count == 3


class TestUsersFindOne:
    @responses.activate
    def test_default(self):
        # validate first result returned
        responses.get(
            "https://connect.example/__api__/v1/users",
            json=load_mock("v1/users?page_number=1&page_size=500.jsonc"),
        )
        con = Client(api_key="12345", url="https://connect.example/")
        user = con.users.find_one()
        assert user
        assert user["username"] == "al"
        assert len(responses.calls) == 1

    @responses.activate
    def test_params(self):
        mock_get = responses.get(
            "https://connect.example/__api__/v1/users",
            match=[
                responses.matchers.query_param_matcher(
                    {
                        "page_size": 500,
                        "page_number": 1,
                        "key1": "value1",
                        "key2": "value2",
                        "key3": "value3",
                    },
                ),
            ],
            json=load_mock("v1/users?page_number=1&page_size=500.jsonc"),
        )
        con = Client(api_key="12345", url="https://connect.example/")
        con.users.find_one(key1="value1", key2="value2", key3="value3")  # pyright: ignore[reportCallIssue]
        assert mock_get.call_count == 1

    @responses.activate
    def test_empty_results(self):
        responses.get(
            "https://connect.example/__api__/v1/users",
            json={"total": 0, "current_page": 1, "results": []},
        )

        con = Client(api_key="12345", url="https://connect.example/")
        user = con.users.find_one()
        assert user is None


class TestUsersFind:
    @responses.activate
    def test_default(self):
        # validate response body is parsed and returned
        responses.get(
            "https://connect.example/__api__/v1/users",
            match=[responses.matchers.query_param_matcher({"page_size": 500, "page_number": 1})],
            json=load_mock("v1/users?page_number=1&page_size=500.jsonc"),
        )
        responses.get(
            "https://connect.example/__api__/v1/users",
            match=[responses.matchers.query_param_matcher({"page_size": 500, "page_number": 2})],
            json=load_mock("v1/users?page_number=2&page_size=500.jsonc"),
        )
        con = Client(api_key="12345", url="https://connect.example/")
        users = con.users.find()
        assert len(users) == 3
        assert users[0] == {
            "email": "alice@connect.example",
            "username": "al",
            "first_name": "Alice",
            "last_name": "User",
            "user_role": "publisher",
            "created_time": "2017-08-08T15:24:32Z",
            "updated_time": "2023-03-02T20:25:06Z",
            "active_time": "2018-05-09T16:58:45Z",
            "confirmed": True,
            "locked": False,
            "guid": "a01792e3-2e67-402e-99af-be04a48da074",
        }

    @responses.activate
    def test_params(self):
        # validate input params are propagated to the query params
        params = {"key1": "value1", "key2": "value2", "key3": "value3"}
        responses.get(
            "https://connect.example/__api__/v1/users",
            match=[
                responses.matchers.query_param_matcher(
                    {
                        "page_size": 500,
                        "page_number": 1,
                        "key1": "value1",
                        "key2": "value2",
                        "key3": "value3",
                    },
                ),
            ],
            json=load_mock("v1/users?page_number=1&page_size=500.jsonc"),
        )
        responses.get(
            "https://connect.example/__api__/v1/users",
            match=[
                responses.matchers.query_param_matcher(
                    {
                        "page_size": 500,
                        "page_number": 2,
                        "key1": "value1",
                        "key2": "value2",
                        "key3": "value3",
                    },
                ),
            ],
            json=load_mock("v1/users?page_number=2&page_size=500.jsonc"),
        )
        con = Client(api_key="12345", url="https://connect.example/")
        con.users.find_one(key1="value1", key2="value2", key3="value3")  # pyright: ignore[reportCallIssue]
        responses.assert_call_count(
            "https://connect.example/__api__/v1/users?key1=value1&key2=value2&key3=value3&page_number=1&page_size=500",
            1,
        )

    @responses.activate
    def test_params_not_dict_like(self):
        # validate input params are propagated to the query params
        con = Client(api_key="12345", url="https://connect.example/")
        not_dict_like = "string"
        with pytest.raises(TypeError):
            con.users.find(
                not_dict_like  # pyright: ignore[reportCallIssue]
            )
